"""

Notes for using the plugin-related fixtures here:

1. The `npe2pm_` fixture is always used, and it mocks the global npe2 plugin
   manager instance with a discovery-deficient plugin manager.  No plugins should be
   discovered in tests without explicit registration.
2. wherever the builtins need to be tested, the `builtins` fixture should be explicitly
   added to the test.  (it's a DynamicPlugin that registers our builtins.yaml with the
   global mock npe2 plugin manager)
3. wherever *additional* plugins or contributions need to be added, use the `tmp_plugin`
   fixture, and add additional contributions _within_ the test (not in the fixture):
    ```python
    def test_something(tmp_plugin):
        @tmp_plugin.contribute.reader(filname_patterns=["*.ext"])
        def f(path): ...

        # the plugin name can be accessed at:
        tmp_plugin.name
    ```
4. If you need a _second_ mock plugin, use `tmp_plugin.spawn(register=True)` to create
   another one.
   ```python
   new_plugin = tmp_plugin.spawn(register=True)

   @new_plugin.contribute.reader(filename_patterns=["*.tiff"])
   def get_reader(path):
       ...
   ```
"""

from __future__ import annotations

import contextlib
import os
import sys
import threading
from concurrent.futures import ThreadPoolExecutor
from contextlib import suppress
from functools import partial
from itertools import chain
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import TYPE_CHECKING, Optional
from weakref import WeakKeyDictionary

from npe2 import PackageMetadata

with suppress(ModuleNotFoundError):
    __import__('dotenv').load_dotenv()

from datetime import timedelta
from time import perf_counter

import dask.threaded
import numpy as np
import pytest
from _pytest.pathlib import bestrelpath
from IPython.core.history import HistoryManager
from packaging.version import parse as parse_version
from pytest_pretty import CustomTerminalReporter

from napari.components import LayerList
from napari.layers import Image, Labels, Points, Shapes, Vectors
from napari.utils.misc import ROOT_DIR

if TYPE_CHECKING:
    from npe2._pytest_plugin import TestPluginManager

# touch ~/.Xauthority for Xlib support, must happen before importing pyautogui
if os.getenv('CI') and sys.platform.startswith('linux'):
    xauth = Path('~/.Xauthority').expanduser()
    if not xauth.exists():
        xauth.touch()


@pytest.fixture
def layer_data_and_types():
    """Fixture that provides some layers and filenames

    Returns
    -------
    tuple
        ``layers, layer_data, layer_types, filenames``

        - layers: some image and points layers
        - layer_data: same as above but in LayerData form
        - layer_types: list of strings with type of layer
        - filenames: the expected filenames with extensions for the layers.
    """
    layers = [
        Image(np.random.rand(20, 20), name='ex_img'),
        Image(np.random.rand(20, 20)),
        Points(np.random.rand(20, 2), name='ex_pts'),
        Points(
            np.random.rand(20, 2), properties={'values': np.random.rand(20)}
        ),
    ]
    extensions = ['.tif', '.tif', '.csv', '.csv']
    layer_data = [layer.as_layer_data_tuple() for layer in layers]
    layer_types = [layer._type_string for layer in layers]
    filenames = [layer.name + e for layer, e in zip(layers, extensions)]
    return layers, layer_data, layer_types, filenames


@pytest.fixture(
    params=[
        'image',
        'labels',
        'points',
        'shapes',
        'shapes-rectangles',
        'vectors',
    ]
)
def layer(request):
    """Parameterized fixture that supplies a layer for testing.

    Parameters
    ----------
    request : _pytest.fixtures.SubRequest
        The pytest request object

    Returns
    -------
    napari.layers.Layer
        The desired napari Layer.
    """
    np.random.seed(0)
    if request.param == 'image':
        data = np.random.rand(20, 20)
        return Image(data)
    if request.param == 'labels':
        data = np.random.randint(10, size=(20, 20))
        return Labels(data)
    if request.param == 'points':
        data = np.random.rand(20, 2)
        return Points(data)
    if request.param == 'shapes':
        data = [
            np.random.rand(2, 2),
            np.random.rand(2, 2),
            np.random.rand(6, 2),
            np.random.rand(6, 2),
            np.random.rand(2, 2),
        ]
        shape_type = ['ellipse', 'line', 'path', 'polygon', 'rectangle']
        return Shapes(data, shape_type=shape_type)
    if request.param == 'shapes-rectangles':
        data = np.random.rand(7, 4, 2)
        return Shapes(data)
    if request.param == 'vectors':
        data = np.random.rand(20, 2, 2)
        return Vectors(data)

    return None


@pytest.fixture
def layers():
    """Fixture that supplies a layers list for testing.

    Returns
    -------
    napari.components.LayerList
        The desired napari LayerList.
    """
    np.random.seed(0)
    list_of_layers = [
        Image(np.random.rand(20, 20)),
        Labels(np.random.randint(10, size=(20, 2))),
        Points(np.random.rand(20, 2)),
        Shapes(np.random.rand(10, 2, 2)),
        Vectors(np.random.rand(10, 2, 2)),
    ]
    return LayerList(list_of_layers)


@pytest.fixture(autouse=True)
def _skip_examples(request):
    """Skip examples test if ."""
    if request.node.get_closest_marker(
        'examples'
    ) and request.config.getoption('--skip_examples'):
        pytest.skip('running with --skip_examples')


# _PYTEST_RAISE=1 will prevent pytest from handling exceptions.
# Use with a debugger that's set to break on "unhandled exceptions".
# https://github.com/pytest-dev/pytest/issues/7409
if os.getenv('_PYTEST_RAISE', '0') != '0':

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call):
        raise call.excinfo.value

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo):
        raise excinfo.value


@pytest.fixture(autouse=True)
def _fresh_settings(monkeypatch):
    """This fixture ensures that default settings are used for every test.

    and ensures that changes to settings in a test are reverted, and never
    saved to disk.
    """
    from napari import settings
    from napari.settings import NapariSettings

    # prevent the developer's config file from being used if it exists
    cp = NapariSettings.__private_attributes__['_config_path']
    monkeypatch.setattr(cp, 'default', None)

    # calling save() with no config path is normally an error
    # here we just have save() return if called without a valid path
    NapariSettings.__original_save__ = NapariSettings.save

    def _mock_save(self, path=None, **dict_kwargs):
        if not (path or self.config_path):
            return
        NapariSettings.__original_save__(self, path, **dict_kwargs)

    monkeypatch.setattr(NapariSettings, 'save', _mock_save)

    # this makes sure that we start with fresh settings for every test.
    settings._SETTINGS = None
    return


@pytest.fixture(autouse=True)
def _auto_shutdown_dask_threadworkers():
    """
    This automatically shutdown dask thread workers.

    We don't assert the number of threads in unchanged as other things
    modify the number of threads.
    """
    assert dask.threaded.default_pool is None
    try:
        yield
    finally:
        if isinstance(dask.threaded.default_pool, ThreadPool):
            dask.threaded.default_pool.close()
            dask.threaded.default_pool.join()
        elif dask.threaded.default_pool:
            dask.threaded.default_pool.shutdown()
        dask.threaded.default_pool = None


# this is not the proper way to configure IPython, but it's an easy one.
# This will prevent IPython to try to write history on its sql file and do
# everything in memory.
# 1) it saves a thread and
# 2) it can prevent issues with slow or read-only file systems in CI.
HistoryManager.enabled = False


@pytest.fixture
def napari_svg_name():
    """the plugin name changes with npe2 to `napari-svg` from `svg`."""
    from importlib.metadata import version

    if parse_version(version('napari-svg')) < parse_version('0.1.6'):
        return 'svg'

    return 'napari-svg'


@pytest.fixture(autouse=True)
def npe2pm_(npe2pm, monkeypatch):
    """Autouse npe2 & npe1 mock plugin managers with no registered plugins."""
    from napari.plugins import NapariPluginManager

    monkeypatch.setattr(NapariPluginManager, 'discover', lambda *_, **__: None)
    return npe2pm


@pytest.fixture
def builtins(npe2pm_: TestPluginManager):
    with npe2pm_.tmp_plugin(package='napari') as plugin:
        yield plugin


@pytest.fixture
def tmp_plugin(npe2pm_: TestPluginManager):
    with npe2pm_.tmp_plugin() as plugin:
        plugin.manifest.package_metadata = PackageMetadata(  # type: ignore[call-arg]
            version='0.1.0', name='test'
        )
        plugin.manifest.display_name = 'Temp Plugin'
        yield plugin


@pytest.fixture
def viewer_model():
    from napari.components import ViewerModel

    return ViewerModel()


@pytest.fixture
def qt_viewer_(qtbot, viewer_model, monkeypatch):
    from napari._qt.qt_viewer import QtViewer

    viewer = QtViewer(viewer_model)

    original_controls = viewer.__class__.controls.fget
    original_layers = viewer.__class__.layers.fget
    original_layer_buttons = viewer.__class__.layerButtons.fget
    original_viewer_buttons = viewer.__class__.viewerButtons.fget
    original_dock_layer_list = viewer.__class__.dockLayerList.fget
    original_dock_layer_controls = viewer.__class__.dockLayerControls.fget
    original_dock_console = viewer.__class__.dockConsole.fget
    original_dock_performance = viewer.__class__.dockPerformance.fget

    def hide_widget(widget):
        widget.hide()

    def hide_and_clear_qt_viewer(viewer: QtViewer):
        viewer._instances.clear()
        viewer.hide()

    def patched_controls(self):
        if self._controls is None:
            self._controls = original_controls(self)
            qtbot.addWidget(self._controls, before_close_func=hide_widget)
        return self._controls

    def patched_layers(self):
        if self._layers is None:
            self._layers = original_layers(self)
            qtbot.addWidget(self._layers, before_close_func=hide_widget)
        return self._layers

    def patched_layer_buttons(self):
        if self._layersButtons is None:
            self._layersButtons = original_layer_buttons(self)
            qtbot.addWidget(self._layersButtons, before_close_func=hide_widget)
        return self._layersButtons

    def patched_viewer_buttons(self):
        if self._viewerButtons is None:
            self._viewerButtons = original_viewer_buttons(self)
            qtbot.addWidget(self._viewerButtons, before_close_func=hide_widget)
        return self._viewerButtons

    def patched_dock_layer_list(self):
        if self._dockLayerList is None:
            self._dockLayerList = original_dock_layer_list(self)
            qtbot.addWidget(self._dockLayerList, before_close_func=hide_widget)
        return self._dockLayerList

    def patched_dock_layer_controls(self):
        if self._dockLayerControls is None:
            self._dockLayerControls = original_dock_layer_controls(self)
            qtbot.addWidget(
                self._dockLayerControls, before_close_func=hide_widget
            )
        return self._dockLayerControls

    def patched_dock_console(self):
        if self._dockConsole is None:
            self._dockConsole = original_dock_console(self)
            qtbot.addWidget(self._dockConsole, before_close_func=hide_widget)
        return self._dockConsole

    def patched_dock_performance(self):
        if self._dockPerformance is None:
            self._dockPerformance = original_dock_performance(self)
            qtbot.addWidget(
                self._dockPerformance, before_close_func=hide_widget
            )
        return self._dockPerformance

    monkeypatch.setattr(
        viewer.__class__, 'controls', property(patched_controls)
    )
    monkeypatch.setattr(viewer.__class__, 'layers', property(patched_layers))
    monkeypatch.setattr(
        viewer.__class__, 'layerButtons', property(patched_layer_buttons)
    )
    monkeypatch.setattr(
        viewer.__class__, 'viewerButtons', property(patched_viewer_buttons)
    )
    monkeypatch.setattr(
        viewer.__class__, 'dockLayerList', property(patched_dock_layer_list)
    )
    monkeypatch.setattr(
        viewer.__class__,
        'dockLayerControls',
        property(patched_dock_layer_controls),
    )
    monkeypatch.setattr(
        viewer.__class__, 'dockConsole', property(patched_dock_console)
    )
    monkeypatch.setattr(
        viewer.__class__, 'dockPerformance', property(patched_dock_performance)
    )

    qtbot.addWidget(viewer, before_close_func=hide_and_clear_qt_viewer)
    return viewer


@pytest.fixture
def qt_viewer(qt_viewer_):
    """We created `qt_viewer_` fixture to allow modifying qt_viewer
    if module-level-specific modifications are necessary.
    For example, in `test_qt_viewer.py`.
    """
    return qt_viewer_


@pytest.fixture(autouse=True)
def _clear_cached_action_injection():
    """Automatically clear cached property `Action.injected`.

    Allows action manager actions to be injected using current provider/processors
    and dependencies. See #7219 for details.
    To be removed after ActionManager deprecation.
    """
    from napari.utils.action_manager import action_manager

    for action in action_manager._actions.values():
        if 'injected' in action.__dict__:
            del action.__dict__['injected']


def _event_check(instance):
    def _prepare_check(name, no_event_):
        def check(instance, no_event=no_event_):
            if name in no_event:
                assert not hasattr(
                    instance.events, name
                ), f'event {name} defined'
            else:
                assert hasattr(
                    instance.events, name
                ), f'event {name} not defined'

        return check

    no_event_set = set()
    if isinstance(instance, tuple):
        no_event_set = instance[1]
        instance = instance[0]

    for name, value in instance.__class__.__dict__.items():
        if isinstance(value, property) and name[0] != '_':
            yield _prepare_check(name, no_event_set), instance, name


def pytest_generate_tests(metafunc):
    """Generate separate test for each test toc check if all events are defined."""
    if 'event_define_check' in metafunc.fixturenames:
        res = []
        ids = []

        for obj in metafunc.cls.get_objects():
            for check, instance, name in _event_check(obj):
                res.append((check, instance))
                ids.append(f'{name}-{instance}')

        metafunc.parametrize('event_define_check,obj', res, ids=ids)


def pytest_collection_modifyitems(session, config, items):
    test_subset = os.environ.get('NAPARI_TEST_SUBSET')

    test_order_prefix = [
        os.path.join('napari', 'utils'),
        os.path.join('napari', 'layers'),
        os.path.join('napari', 'components'),
        os.path.join('napari', 'settings'),
        os.path.join('napari', 'plugins'),
        os.path.join('napari', '_vispy'),
        os.path.join('napari', '_qt'),
        os.path.join('napari', 'qt'),
        os.path.join('napari', '_tests'),
        os.path.join('napari', '_tests', 'test_examples.py'),
    ]
    test_order = [[] for _ in test_order_prefix]
    test_order.append([])  # for not matching tests
    for item in items:
        if test_subset:
            if test_subset.lower() == 'qt' and 'qapp' not in item.fixturenames:
                # Skip non Qt tests
                continue
            if (
                test_subset.lower() == 'headless'
                and 'qapp' in item.fixturenames
            ):
                # Skip Qt tests
                continue

        index = -1
        for i, prefix in enumerate(test_order_prefix):
            if prefix in str(item.fspath):
                index = i
        test_order[index].append(item)
    items[:] = list(chain(*test_order))


@pytest.fixture(autouse=True)
def _disable_notification_dismiss_timer(monkeypatch):
    """
    This fixture disables starting timer for closing notification
    by setting the value of `NapariQtNotification.DISMISS_AFTER` to 0.

    As Qt timer is realised by thread and keep reference to the object,
    without increase of reference counter object could be garbage collected and
    cause segmentation fault error when Qt (C++) code try to access it without
    checking if Python object exists.

    This fixture is used in all tests because it is possible to call Qt code
    from non Qt test by connection of `NapariQtNotification.show_notification` to
    `NotificationManager` global instance.
    """

    with suppress(ImportError):
        from napari._qt.dialogs.qt_notification import NapariQtNotification

        monkeypatch.setattr(NapariQtNotification, 'DISMISS_AFTER', 0)
        monkeypatch.setattr(NapariQtNotification, 'FADE_IN_RATE', 0)
        monkeypatch.setattr(NapariQtNotification, 'FADE_OUT_RATE', 0)


@pytest.fixture
def single_threaded_executor():
    executor = ThreadPoolExecutor(max_workers=1)
    yield executor
    executor.shutdown()


def _get_calling_stack():  # pragma: no cover
    stack = []
    for i in range(2, sys.getrecursionlimit()):
        try:
            frame = sys._getframe(i)
        except ValueError:
            break
        stack.append(f'{frame.f_code.co_filename}:{frame.f_lineno}')
    return '\n'.join(stack)


def _get_calling_place(depth=1):  # pragma: no cover
    if not hasattr(sys, '_getframe'):
        return ''
    frame = sys._getframe(1 + depth)
    result = f'{frame.f_code.co_filename}:{frame.f_lineno}'
    if not frame.f_code.co_filename.startswith(ROOT_DIR):
        with suppress(ValueError):
            while not frame.f_code.co_filename.startswith(ROOT_DIR):
                frame = frame.f_back
                if frame is None:
                    break
            else:
                result += f' called from\n{frame.f_code.co_filename}:{frame.f_lineno}'
    return result


@pytest.fixture
def _dangling_qthreads(monkeypatch, qtbot, request):
    from qtpy.QtCore import QThread

    base_start = QThread.start
    thread_dict = WeakKeyDictionary()
    base_constructor = QThread.__init__

    def run_with_trace(self):  # pragma: no cover
        """
        QThread.run but adding execution to sys.settrace when measuring coverage.

        See https://github.com/nedbat/coveragepy/issues/686#issuecomment-634932753
        and `init_with_trace`. When running QThreads during testing, we monkeypatch
        the QThread constructor and run methods with traceable equivalents.
        """
        if 'coverage' in sys.modules:
            # https://github.com/nedbat/coveragepy/issues/686#issuecomment-634932753
            sys.settrace(threading._trace_hook)
        self._base_run()

    def init_with_trace(self, *args, **kwargs):
        """Constructor for QThread adding tracing for coverage measurements.

        Functions running in QThreads don't get measured by coverage.py, see
        https://github.com/nedbat/coveragepy/issues/686. Therefore, we will
        monkeypatch the constructor to add to the thread to `sys.settrace` when
        we call `run` and `coverage` is in `sys.modules`.
        """
        base_constructor(self, *args, **kwargs)
        self._base_run = self.run
        self.run = partial(run_with_trace, self)

    # dict of threads that have been started but not yet terminated

    if 'disable_qthread_start' in request.keywords:

        def start_with_save_reference(self, priority=QThread.InheritPriority):
            """Dummy function to prevent thread starts."""

    else:

        def start_with_save_reference(self, priority=QThread.InheritPriority):
            """Thread start function with logs to detect hanging threads.

            Saves a weak reference to the thread and detects hanging threads,
            as well as where the threads were started.
            """
            thread_dict[self] = _get_calling_place()
            base_start(self, priority)

    monkeypatch.setattr(QThread, 'start', start_with_save_reference)
    monkeypatch.setattr(QThread, '__init__', init_with_trace)

    yield

    dangling_threads_li = []

    for thread, calling in thread_dict.items():
        try:
            if thread.isRunning():
                dangling_threads_li.append((thread, calling))
        except RuntimeError as e:
            if (
                'wrapped C/C++ object of type' not in e.args[0]
                and 'Internal C++ object' not in e.args[0]
            ):
                raise

    for thread, _ in dangling_threads_li:
        with suppress(RuntimeError):
            thread.quit()
            qtbot.waitUntil(thread.isFinished, timeout=2000)

    long_desc = (
        'If you see this error, it means that a QThread was started in a test '
        'but not terminated. This can cause segfaults in the test suite. '
        'Please use the `qtbot` fixture to wait for the thread to finish. '
        'If you think that the thread is obsolete for this test, you can '
        'use the `@pytest.mark.disable_qthread_start` mark or  `monkeypatch` '
        'fixture to patch the `start` method of the '
        'QThread class to do nothing.\n'
    )

    if len(dangling_threads_li) > 1:
        long_desc += ' The QThreads were started in:\n'
    else:
        long_desc += ' The QThread was started in:\n'

    assert not dangling_threads_li, long_desc + '\n'.join(
        x[1] for x in dangling_threads_li
    )


@pytest.fixture
def _dangling_qthread_pool(monkeypatch, request):
    from qtpy.QtCore import QThreadPool

    base_start = QThreadPool.start
    threadpool_dict = WeakKeyDictionary()
    # dict of threadpools that have been used to run QRunnables

    if 'disable_qthread_pool_start' in request.keywords:

        def my_start(self, runnable, priority=0):
            """dummy function to prevent thread start"""

    else:

        def my_start(self, runnable, priority=0):
            if self not in threadpool_dict:
                threadpool_dict[self] = []
            threadpool_dict[self].append(_get_calling_place())
            base_start(self, runnable, priority)

    monkeypatch.setattr(QThreadPool, 'start', my_start)
    yield

    dangling_threads_pools = []

    for thread_pool, calling in threadpool_dict.items():
        thread_pool.clear()
        thread_pool.waitForDone(20)
        if thread_pool.activeThreadCount():
            dangling_threads_pools.append((thread_pool, calling))

    for thread_pool, _ in dangling_threads_pools:
        with suppress(RuntimeError):
            thread_pool.clear()
            thread_pool.waitForDone(2000)

    long_desc = (
        'If you see this error, it means that a QThreadPool was used to run '
        'a QRunnable in a test but not terminated. This can cause segfaults '
        'in the test suite. Please use the `qtbot` fixture to wait for the '
        'thread to finish. If you think that the thread is obsolete for this '
        'use the `@pytest.mark.disable_qthread_pool_start` mark or  `monkeypatch` '
        'fixture to patch the `start` '
        'method of the QThreadPool class to do nothing.\n'
    )
    if len(dangling_threads_pools) > 1:
        long_desc += ' The QThreadPools were used in:\n'
    else:
        long_desc += ' The QThreadPool was used in:\n'

    assert not dangling_threads_pools, long_desc + '\n'.join(
        '; '.join(x[1]) for x in dangling_threads_pools
    )


@pytest.fixture
def _dangling_qtimers(monkeypatch, request):
    from qtpy.QtCore import QTimer

    base_start = QTimer.start
    timer_dkt = WeakKeyDictionary()
    single_shot_list = []

    if 'disable_qtimer_start' in request.keywords:
        from pytestqt.qt_compat import qt_api

        def my_start(self, msec=None):
            """dummy function to prevent timer start"""

        _single_shot = my_start

        class OldTimer(QTimer):
            def start(self, time=None):
                if time is not None:
                    base_start(self, time)
                else:
                    base_start(self)

        monkeypatch.setattr(qt_api.QtCore, 'QTimer', OldTimer)
        # This monkeypatch is require to keep `qtbot.waitUntil` working

    else:

        def my_start(self, msec=None):
            calling_place = _get_calling_place()
            if 'superqt' in calling_place and 'throttler' in calling_place:
                calling_place += f' - {_get_calling_place(2)}'
            timer_dkt[self] = calling_place
            if msec is not None:
                base_start(self, msec)
            else:
                base_start(self)

        def single_shot(msec, reciver, method=None):
            t = QTimer()
            t.setSingleShot(True)
            if method is None:
                t.timeout.connect(reciver)
            else:
                t.timeout.connect(getattr(reciver, method))
            calling_place = _get_calling_place(2)
            if 'superqt' in calling_place and 'throttler' in calling_place:
                calling_place += _get_calling_stack()
            single_shot_list.append((t, _get_calling_place(2)))
            base_start(t, msec)

        def _single_shot(self, *args):
            if isinstance(self, QTimer):
                single_shot(*args)
            else:
                single_shot(self, *args)

    monkeypatch.setattr(QTimer, 'start', my_start)
    monkeypatch.setattr(QTimer, 'singleShot', _single_shot)

    yield

    dangling_timers = []

    for timer, calling in chain(timer_dkt.items(), single_shot_list):
        if timer.isActive():
            dangling_timers.append((timer, calling))

    for timer, _ in dangling_timers:
        with suppress(RuntimeError):
            timer.stop()

    long_desc = (
        'If you see this error, it means that a QTimer was started but not stopped. '
        'This can cause tests to fail, and can also cause segfaults. '
        'If this test does not require a QTimer to pass you could monkeypatch it out. '
        'If it does require a QTimer, you should stop or wait for it to finish before test ends. '
    )
    if len(dangling_timers) > 1:
        long_desc += 'The QTimers were started in:\n'
    else:
        long_desc += 'The QTimer was started in:\n'

    def _check_throttle_info(path):
        if 'superqt' in path and 'throttler' in path:
            return (
                path
                + " it's possible that there was a problem with unfinished work by a "
                'qthrottler; to solve this, you can either try to wait (such as with '
                '`qtbot.wait`) or disable throttling with the disable_throttling fixture'
            )
        return path

    assert not dangling_timers, long_desc + '\n'.join(
        _check_throttle_info(x[1]) for x in dangling_timers
    )


def _throttle_mock(self):
    self.triggered.emit()


def _flush_mock(self):
    """There are no waiting events."""


@pytest.fixture
def _disable_throttling(monkeypatch):
    """Disable qthrottler from superqt.

    This is sometimes necessary to avoid flaky failures in tests
    due to dangling qt timers.
    """
    # if this monkeypath fails then you should update path to GenericSignalThrottler
    monkeypatch.setattr(
        'superqt.utils._throttler.GenericSignalThrottler.throttle',
        _throttle_mock,
    )
    monkeypatch.setattr(
        'superqt.utils._throttler.GenericSignalThrottler.flush', _flush_mock
    )


@pytest.fixture
def _dangling_qanimations(monkeypatch, request):
    from qtpy.QtCore import QPropertyAnimation

    base_start = QPropertyAnimation.start
    animation_dkt = WeakKeyDictionary()

    if 'disable_qanimation_start' in request.keywords:

        def my_start(self):
            """dummy function to prevent thread start"""

    else:

        def my_start(self):
            animation_dkt[self] = _get_calling_place()
            base_start(self)

    monkeypatch.setattr(QPropertyAnimation, 'start', my_start)
    yield

    dangling_animations = []

    for animation, calling in animation_dkt.items():
        if animation.state() == QPropertyAnimation.Running:
            dangling_animations.append((animation, calling))

    for animation, _ in dangling_animations:
        with suppress(RuntimeError):
            animation.stop()

    long_desc = (
        'If you see this error, it means that a QPropertyAnimation was started but not stopped. '
        'This can cause tests to fail, and can also cause segfaults. '
        'If this test does not require a QPropertyAnimation to pass you could monkeypatch it out. '
        'If it does require a QPropertyAnimation, you should stop or wait for it to finish before test ends. '
    )
    if len(dangling_animations) > 1:
        long_desc += ' The QPropertyAnimations were started in:\n'
    else:
        long_desc += ' The QPropertyAnimation was started in:\n'
    assert not dangling_animations, long_desc + '\n'.join(
        x[1] for x in dangling_animations
    )


with contextlib.suppress(ImportError):
    # in headless test suite we don't have Qt bindings
    # So we cannot inherit from QtBot and declare the fixture

    from pytestqt.qtbot import QtBot

    class QtBotWithOnCloseRenaming(QtBot):
        """Modified QtBot that renames widgets when closing them in tests.

        After a test ends that uses QtBot, all instantiated widgets added to
        the bot have their name changed to 'handled_widget'. This allows us to
        detect leaking widgets at the end of a test run, and avoid the
        segmentation faults that often result from such leaks. [1]_

        See Also
        --------
        `_find_dangling_widgets`: fixture that finds all widgets that have not
        been renamed to 'handled_widget'.

        References
        ----------
        .. [1] https://czaki.github.io/blog/2024/09/16/preventing-segfaults-in-test-suite-that-has-qt-tests/
        """

        def addWidget(self, widget, *, before_close_func=None):
            if widget.objectName() == '':
                # object does not have a name, so we can set it
                widget.setObjectName('handled_widget')
                before_close_func_ = before_close_func
            elif before_close_func is None:
                # there is no custom teardown function,
                # so we provide one that will set object name

                def before_close_func_(w):
                    w.setObjectName('handled_widget')
            else:
                # user provided custom teardown function,
                # so we need to wrap it to set object name

                def before_close_func_(w):
                    before_close_func(w)
                    w.setObjectName('handled_widget')

            super().addWidget(widget, before_close_func=before_close_func_)

    @pytest.fixture
    def qtbot(qapp, request):  # pragma: no cover
        """Fixture to create a QtBotWithOnCloseRenaming instance for testing.

        Make sure to call addWidget for each top-level widget you create to
        ensure that they are properly closed after the test ends.

        The `qapp` fixture is used to ensure that the QApplication is created
        before, so we need it, even without using it directly in this fixture.
        """
        return QtBotWithOnCloseRenaming(request)


@pytest.fixture
def _find_dangling_widgets(request, qtbot):
    yield

    from qtpy.QtWidgets import QApplication

    from napari._qt.qt_main_window import _QtMainWindow

    top_level_widgets = QApplication.topLevelWidgets()

    viewer_weak_set = getattr(request.node, '_viewer_weak_set', set())

    problematic_widgets = []

    for widget in top_level_widgets:
        if widget.parent() is not None:
            continue
        if (
            isinstance(widget, _QtMainWindow)
            and widget._qt_viewer.viewer in viewer_weak_set
        ):
            continue

        if widget.__class__.__module__.startswith('qtconsole'):
            continue

        if widget.objectName() == 'handled_widget':
            continue

        if widget.__class__.__name__ == 'CanvasBackendDesktop':
            # TODO: we don't understand why this class leaks in
            #  napari/_tests/test_sys_info.py, so we make an exception
            #  here and we don't raise when this class leaks.
            continue

        problematic_widgets.append(widget)

    if problematic_widgets:
        text = '\n'.join(
            f'Widget: {widget} of type {type(widget)} with name {widget.objectName()}'
            for widget in problematic_widgets
        )

        for widget in problematic_widgets:
            widget.setObjectName('handled_widget')

        raise RuntimeError(f'Found dangling widgets:\n{text}')


def pytest_runtest_setup(item):
    """Add Qt leak detection fixtures *only* in tests using the qapp fixture.

    Because we have headless test suite that does not include Qt, we cannot
    simply use `@pytest.fixture(autouse=True)` on all our fixtures for
    detecting leaking Qt objects.

    Instead, here we detect whether the `qapp` fixture is being used, detecting
    tests that use Qt and need to be checked for Qt objects leaks.

    A note to maintainers: tests *may* attempt to use Qt classes but not use
    the `qapp` fixture. This is BAD, and may cause Qt failures to be reported
    far away from the problematic code or test. If you find any tests
    instantiating Qt objects but not using qapp or qtbot, please submit a PR
    adding the qtbot fixture and adding any top-level Qt widgets with::

        qtbot.addWidget(widget_instance)

    """

    if 'qapp' in item.fixturenames:
        # here we do autouse for dangling fixtures only if qapp is used
        if 'qtbot' not in item.fixturenames:
            # for proper waiting for threads to finish
            item.fixturenames.append('qtbot')
        item.fixturenames.extend(
            [
                '_find_dangling_widgets',
                '_dangling_qthread_pool',
                '_dangling_qanimations',
                '_dangling_qthreads',
                '_dangling_qtimers',
            ]
        )


class NapariTerminalReporter(CustomTerminalReporter):
    """
    This ia s custom terminal reporter to how long it takes to finish given part of tests.
    It prints time each time when test from different file is started.

    It is created to be able to see if timeout is caused by long time execution, or it is just hanging.
    """

    currentfspath: Optional[Path]

    def write_fspath_result(self, nodeid: str, res, **markup: bool) -> None:
        if getattr(self, '_start_time', None) is None:
            self._start_time = perf_counter()
        fspath = self.config.rootpath / nodeid.split('::')[0]
        if self.currentfspath is None or fspath != self.currentfspath:
            if self.currentfspath is not None and self._show_progress_info:
                self._write_progress_information_filling_space()
                if os.environ.get('CI', False):
                    self.write(
                        f' [{timedelta(seconds=int(perf_counter() - self._start_time))}]'
                    )
            self.currentfspath = fspath
            relfspath = bestrelpath(self.startpath, fspath)
            self._tw.line()
            self.write(relfspath + ' ')
        self.write(res, flush=True, **markup)


@pytest.hookimpl(trylast=True)
def pytest_configure(config):
    # Get the standard terminal reporter plugin and replace it with our
    standard_reporter = config.pluginmanager.getplugin('terminalreporter')
    custom_reporter = NapariTerminalReporter(config, sys.stdout)
    if standard_reporter._session is not None:
        custom_reporter._session = standard_reporter._session
    config.pluginmanager.unregister(standard_reporter)
    config.pluginmanager.register(custom_reporter, 'terminalreporter')



from functools import lru_cache

import numpy as np
from scipy import ndimage as ndi


def interpolate_coordinates(old_coord, new_coord, brush_size):
    """Interpolates coordinates depending on brush size.

    Useful for ensuring painting is continuous in labels layer.

    Parameters
    ----------
    old_coord : np.ndarray, 1x2
        Last position of cursor.
    new_coord : np.ndarray, 1x2
        Current position of cursor.
    brush_size : float
        Size of brush, which determines spacing of interpolation.

    Returns
    -------
    coords : np.array, Nx2
        List of coordinates to ensure painting is continuous
    """
    if old_coord is None:
        old_coord = new_coord
    if new_coord is None:
        new_coord = old_coord
    num_step = round(
        max(abs(np.array(new_coord) - np.array(old_coord))) / brush_size * 4
    )
    coords = [
        np.linspace(old_coord[i], new_coord[i], num=int(num_step + 1))
        for i in range(len(new_coord))
    ]
    coords = np.stack(coords).T
    if len(coords) > 1:
        coords = coords[1:]

    return coords


@lru_cache(maxsize=64)
def sphere_indices(radius, scale):
    """Generate centered indices within circle or n-dim ellipsoid.

    Parameters
    ----------
    radius : float
        Radius of circle/sphere
    scale : tuple of float
        The scaling to apply to the sphere along each axis

    Returns
    -------
    mask_indices : array
        Centered indices within circle/sphere
    """
    ndim = len(scale)
    abs_scale = np.abs(scale)
    scale_normalized = np.asarray(abs_scale, dtype=float) / np.min(abs_scale)
    # Create multi-dimensional grid to check for
    # circle/membership around center
    r_normalized = radius / scale_normalized + 0.5
    slices = [
        slice(-int(np.ceil(r)), int(np.floor(r)) + 1) for r in r_normalized
    ]

    indices = np.mgrid[slices].T.reshape(-1, ndim)
    distances_sq = np.sum((indices * scale_normalized) ** 2, axis=1)
    # Use distances within desired radius to mask indices in grid
    mask_indices = indices[distances_sq <= radius**2].astype(int)

    return mask_indices


def indices_in_shape(idxs, shape):
    """Return idxs after filtering out indices that are not in given shape.

    Parameters
    ----------
    idxs : tuple of array of int, or 2D array of int
        The input coordinates. These should be in one of two formats:

        - a tuple of 1D arrays, as for NumPy fancy indexing, or
        - a 2D array of shape (ncoords, ndim), as a list of coordinates

    shape : tuple of int
        The shape in which all indices must fit.

    Returns
    -------
    idxs_filtered : tuple of array of int, or 2D array of int
        The subset of the input idxs that falls within shape.

    Examples
    --------
    >>> idxs0 = (np.array([5, 45, 2]), np.array([6, 5, -5]))
    >>> indices_in_shape(idxs0, (10, 10))
    (array([5]), array([6]))
    >>> idxs1 = np.transpose(idxs0)
    >>> indices_in_shape(idxs1, (10, 10))
    array([[5, 6]])
    """
    np_index = isinstance(idxs, tuple)
    if np_index:  # normalize to 2D coords array
        idxs = np.transpose(idxs)
    keep_coords = np.logical_and(
        np.all(idxs >= 0, axis=1), np.all(idxs < np.array(shape), axis=1)
    )
    filtered = idxs[keep_coords]
    if np_index:  # convert back to original format
        filtered = tuple(filtered.T)
    return filtered


def get_dtype(layer):
    """Returns dtype of layer data

    Parameters
    ----------
    layer : Labels
        Labels layer (may be multiscale)

    Returns
    -------
    dtype
        dtype of Layer data
    """
    layer_data = layer.data
    if not isinstance(layer_data, list):
        layer_data = [layer_data]
    layer_data_level = layer_data[0]
    if hasattr(layer_data_level, 'dtype'):
        layer_dtype = layer_data_level[0].dtype
    else:
        layer_dtype = type(layer_data_level)

    return layer_dtype


def first_nonzero_coordinate(data, start_point, end_point):
    """Coordinate of the first nonzero element between start and end points.

    Parameters
    ----------
    data : nD array, shape (N1, N2, ..., ND)
        A data volume.
    start_point : array, shape (D,)
        The start coordinate to check.
    end_point : array, shape (D,)
        The end coordinate to check.

    Returns
    -------
    coordinates : array of int, shape (D,)
        The coordinates of the first nonzero element along the ray, or None.
    """
    shape = np.asarray(data.shape)
    length = np.linalg.norm(end_point - start_point)
    length_int = np.round(length).astype(int)
    coords = np.linspace(start_point, end_point, length_int + 1, endpoint=True)
    clipped_coords = np.clip(np.round(coords), 0, shape - 1).astype(int)
    nonzero = np.flatnonzero(data[tuple(clipped_coords.T)])
    return None if len(nonzero) == 0 else clipped_coords[nonzero[0]]


def mouse_event_to_labels_coordinate(layer, event):
    """Return the data coordinate of a Labels layer mouse event in 2D or 3D.

    In 2D, this is just the event's position transformed by the layer's
    world_to_data transform.

    In 3D, a ray is cast in data coordinates, and the coordinate of the first
    nonzero value along that ray is returned. If the ray only contains zeros,
    None is returned.

    Parameters
    ----------
    layer : napari.layers.Labels
        The Labels layer.
    event : vispy MouseEvent
        The mouse event, containing position and view direction attributes.

    Returns
    -------
    coordinates : array of int or None
        The data coordinates for the mouse event.
    """
    ndim = len(layer._slice_input.displayed)
    if ndim == 2:
        coordinates = layer.world_to_data(event.position)
    else:  # 3d
        start, end = layer.get_ray_intersections(
            position=event.position,
            view_direction=event.view_direction,
            dims_displayed=layer._slice_input.displayed,
            world=True,
        )
        if start is None and end is None:
            return None
        coordinates = first_nonzero_coordinate(layer.data, start, end)
    return coordinates


def get_contours(labels: np.ndarray, thickness: int, background_label: int):
    """Computes the contours of a 2D label image.

    Parameters
    ----------
    labels : array of integers
        An input labels image.
    thickness : int
        It controls the thickness of the inner boundaries. The outside thickness is always 1.
        The final thickness of the contours will be `thickness + 1`.
    background_label : int
        That label is used to fill everything outside the boundaries.

    Returns
    -------
    A new label image in which only the boundaries of the input image are kept.
    """
    struct_elem = ndi.generate_binary_structure(labels.ndim, 1)

    thick_struct_elem = ndi.iterate_structure(struct_elem, thickness).astype(
        bool
    )

    dilated_labels = ndi.grey_dilation(labels, footprint=struct_elem)
    eroded_labels = ndi.grey_erosion(labels, footprint=thick_struct_elem)
    not_boundaries = dilated_labels == eroded_labels

    contours = labels.copy()
    contours[not_boundaries] = background_label

    return contours


def expand_slice(
    axes_slice: tuple[slice, ...], shape: tuple, offset: int
) -> tuple[slice, ...]:
    """Expands or shrinks a provided multi-axis slice by a given offset"""
    return tuple(
        slice(
            max(0, min(max_size, s.start - offset)),
            max(0, min(max_size, s.stop + offset)),
            s.step,
        )
        for s, max_size in zip(axes_slice, shape)
    )



from unittest.mock import MagicMock, patch

import numpy as np
import pytest
from napari_plugin_engine import HookImplementation

from napari._tests.utils import layer_test_data
from napari.components.viewer_model import ViewerModel
from napari.layers._source import Source

img = np.random.rand(10, 10)
layer_data = [(lay[1], {}, lay[0].__name__.lower()) for lay in layer_test_data]


def _impl(path):
    """just a dummy Hookimpl object to return from mocks"""


_testimpl = HookImplementation(_impl, plugin_name='testimpl')


@pytest.mark.parametrize('layer_datum', layer_data)
def test_add_layers_with_plugins(layer_datum):
    """Test that add_layers_with_plugins adds the expected layer types."""
    with patch(
        'napari.plugins.io.read_data_with_plugins',
        MagicMock(return_value=([layer_datum], _testimpl)),
    ):
        v = ViewerModel()
        v._add_layers_with_plugins(['mock_path'], stack=False)
        layertypes = [layer._type_string for layer in v.layers]
        assert layertypes == [layer_datum[2]]

        expected_source = Source(path='mock_path', reader_plugin='testimpl')
        assert all(lay.source == expected_source for lay in v.layers)


@patch(
    'napari.plugins.io.read_data_with_plugins',
    MagicMock(return_value=([], _testimpl)),
)
def test_plugin_returns_nothing():
    """Test that a plugin returning nothing adds nothing to the Viewer."""
    v = ViewerModel()
    v._add_layers_with_plugins(['mock_path'], stack=False)
    assert not v.layers


@patch(
    'napari.plugins.io.read_data_with_plugins',
    MagicMock(return_value=([(img,)], _testimpl)),
)
def test_viewer_open():
    """Test that a plugin to returning an image adds stuff to the viewer."""
    viewer = ViewerModel()
    assert len(viewer.layers) == 0
    viewer.open('mock_path.tif')
    assert len(viewer.layers) == 1
    # The name should be taken from the path name, stripped of extension
    assert viewer.layers[0].name == 'mock_path'

    # stack=True also works... and very long names are truncated
    viewer.open('mock_path.tif', stack=True)
    assert len(viewer.layers) == 2
    assert viewer.layers[1].name.startswith('mock_path')

    expected_source = Source(path='mock_path.tif', reader_plugin='testimpl')
    assert all(lay.source == expected_source for lay in viewer.layers)

    viewer.open([], stack=[], plugin=None)
    assert len(viewer.layers) == 2


def test_viewer_open_no_plugin(tmp_path):
    viewer = ViewerModel()
    fname = tmp_path / 'gibberish.gbrsh'
    fname.touch()
    with pytest.raises(ValueError, match='.*gibberish.gbrsh.*'):
        # will default to builtins
        viewer.open(fname)


plugin_returns = [
    ([(img, {'name': 'foo'})], {'name': 'bar'}),
    ([(img, {'blending': 'additive'}), (img,)], {'blending': 'translucent'}),
]


@pytest.mark.parametrize(('layer_data', 'kwargs'), plugin_returns)
def test_add_layers_with_plugins_and_kwargs(layer_data, kwargs):
    """Test that _add_layers_with_plugins kwargs override plugin kwargs.

    see also: napari.components._test.test_prune_kwargs
    """
    with patch(
        'napari.plugins.io.read_data_with_plugins',
        MagicMock(return_value=(layer_data, _testimpl)),
    ):
        v = ViewerModel()
        v._add_layers_with_plugins(['mock_path'], kwargs=kwargs, stack=False)
        expected_source = Source(path='mock_path', reader_plugin='testimpl')
        for layer in v.layers:
            for key, val in kwargs.items():
                assert getattr(layer, key) == val
                # if plugins don't provide "name", it falls back to path name
                if 'name' not in kwargs:
                    assert layer.name.startswith('mock_path')
            assert layer.source == expected_source


def test_add_points_layer_with_different_range_updates_all_slices():
    """See https://github.com/napari/napari/pull/4819"""
    viewer = ViewerModel()

    # Adding the first point should show the point
    initial_point = viewer.add_points([[10, 5, 5]])
    np.testing.assert_array_equal(initial_point._indices_view, [0])
    assert viewer.dims.point == (10, 5, 5)

    # Adding an earlier point should keep the dim slider at the position
    # and therefore should not change the viewport.
    earlier_point = viewer.add_points([[8, 1, 1]])
    np.testing.assert_array_equal(initial_point._indices_view, [0])
    np.testing.assert_array_equal(earlier_point._indices_view, [])
    assert viewer.dims.point == (10, 5, 5)

    # Adding a point on the same slice as the initial point should keep the
    # dim slider at the position and should additionally show the added point
    # in the viewport.
    same_slice_as_initial_point = viewer.add_points([[10, 1, 1]])
    np.testing.assert_array_equal(initial_point._indices_view, [0])
    np.testing.assert_array_equal(earlier_point._indices_view, [])
    np.testing.assert_array_equal(
        same_slice_as_initial_point._indices_view, [0]
    )
    assert viewer.dims.point == (10, 5, 5)

    # Adding a later point should keep the dim slider at the position
    # and therefore should not change the viewport.
    later_point = viewer.add_points([[14, 1, 1]])
    np.testing.assert_array_equal(initial_point._indices_view, [0])
    np.testing.assert_array_equal(earlier_point._indices_view, [])
    np.testing.assert_array_equal(
        same_slice_as_initial_point._indices_view, [0]
    )
    np.testing.assert_array_equal(later_point._indices_view, [])
    assert viewer.dims.point == (10, 5, 5)

    # Removing the earlier point should keep the dim slider at the position
    # and therefore should not change the viewport.
    viewer.layers.remove(earlier_point)
    np.testing.assert_array_equal(initial_point._indices_view, [0])
    np.testing.assert_array_equal(
        same_slice_as_initial_point._indices_view, [0]
    )
    np.testing.assert_array_equal(later_point._indices_view, [])
    assert viewer.dims.point == (10, 5, 5)

    # Removing the point on the same slice as the initial point should keep
    # the dim slider at the position and should additionally remove the added
    # point from the viewport.
    viewer.layers.remove(same_slice_as_initial_point)
    np.testing.assert_array_equal(initial_point._indices_view, [0])
    np.testing.assert_array_equal(later_point._indices_view, [])
    assert viewer.dims.point == (10, 5, 5)

    # Removing the initial point should move the dim slider to the later
    # position and update the viewport.
    viewer.layers.remove(initial_point)
    np.testing.assert_array_equal(later_point._indices_view, [0])
    assert viewer.dims.point == (14, 1, 1)

    # Adding an earlier point should keep the dim slider at the position
    # and therefore should not change the viewport.
    earlier_point2 = viewer.add_points([[8, 0, 0]])
    np.testing.assert_array_equal(initial_point._indices_view, [0])
    np.testing.assert_array_equal(earlier_point._indices_view, [])
    assert viewer.dims.point == (14, 1, 1)

    # Removing the second earlier point should move the dim slider to the
    # later position and update the viewport.
    viewer.layers.remove(later_point)
    np.testing.assert_array_equal(earlier_point2._indices_view, [0])
    assert viewer.dims.point == (8, 0, 0)

    # Removing all points should reset the viewport.
    viewer.layers.remove(earlier_point2)
    assert viewer.dims.point == (0, 0)


@pytest.mark.xfail(reason='https://github.com/napari/napari/issues/6198')
def test_last_point_is_visible_in_viewport():
    viewer = ViewerModel()

    # Removing the last point while viewing it should cause
    # us to view the first point due to the layer's new extent.
    points = viewer.add_points([[0, 1, 1], [1, 2, 2]])
    viewer.dims.set_point(0, 1)
    assert viewer.dims.point[0] == 1
    np.testing.assert_array_equal(points._indices_view, [1])

    points.data = [[0, 1, 1]]

    assert viewer.dims.point[0] == 0
    np.testing.assert_array_equal(points._indices_view, [0])
    viewer.layers.remove(points)

    # Removing the first point while viewing it should cause us
    # to view the last point due to the layer's new extent.
    points = viewer.add_points([[0, 1, 1], [1, 2, 2]])
    viewer.dims.set_point(0, 0)
    assert viewer.dims.point[0] == 0
    np.testing.assert_array_equal(points._indices_view, [0])

    points.data = [[1, 2, 2]]

    assert viewer.dims.point[0] == 1
    np.testing.assert_array_equal(points._indices_view, [0])


@pytest.mark.xfail(reason='https://github.com/napari/napari/issues/6199')
def test_dimension_change_is_visible_in_viewport():
    viewer = ViewerModel()

    # Adding a 4d point leads to a visible 4d point with dims.point
    # having the same values.
    point_4d = viewer.add_points([[0] * 4])
    assert viewer.dims.point == tuple([0] * 4)
    np.testing.assert_array_equal(point_4d._indices_view, [0])

    # Adding a 5d point with different 4d coordinates does not change the viewport.
    # Only the first (actual 5th) dimension of the dims.point should change.
    point_5d = viewer.add_points([[2] * 5])
    assert viewer.dims.point == tuple([2] + [0] * 4)
    np.testing.assert_array_equal(point_4d._indices_view, [0])
    np.testing.assert_array_equal(point_5d._indices_view, [])

    # Removing the 4d point leads to an update of the viewport and dims.
    viewer.layers.remove(point_4d)
    assert viewer.dims.point == tuple([2] * 5)
    np.testing.assert_array_equal(point_5d._indices_view, [0])

    # Adding another 4d point does not lead to an update of the viewport
    # because the current dims.point is still in the unified extent.
    point_4d = viewer.add_points([[0] * 4])
    assert viewer.dims.point == tuple([2] * 5)
    np.testing.assert_array_equal(point_4d._indices_view, [])
    np.testing.assert_array_equal(point_5d._indices_view, [0])

    # Removing the 5d point leads to an update of the viewport and dims.
    viewer.layers.remove(point_5d)
    assert viewer.dims.point == tuple([0] * 4)
    np.testing.assert_array_equal(point_4d._indices_view, [0])



import hashlib
import os
import sys
from functools import partial
from typing import Callable

import appdirs

PREFIX_PATH = os.path.realpath(sys.prefix)

sha_short = f'{os.path.basename(PREFIX_PATH)}_{hashlib.sha1(PREFIX_PATH.encode()).hexdigest()}'

_appname = 'napari'
_appauthor = False


# all of these also take an optional "version" argument ... but if we want
# to be able to update napari while using data (e.g. plugins, settings) from
# an earlier version, we should leave off the version.

user_data_dir: Callable[[], str] = partial(
    appdirs.user_data_dir, _appname, _appauthor
)
user_config_dir: Callable[[], str] = partial(
    appdirs.user_config_dir, _appname, _appauthor, sha_short
)
user_cache_dir: Callable[[], str] = partial(
    appdirs.user_cache_dir, _appname, _appauthor, sha_short
)
user_state_dir: Callable[[], str] = partial(
    appdirs.user_state_dir, _appname, _appauthor
)
user_log_dir: Callable[[], str] = partial(
    appdirs.user_log_dir, _appname, _appauthor
)



from napari.utils.events.containers._dict import TypedMutableMapping
from napari.utils.events.containers._evented_dict import EventedDict
from napari.utils.events.containers._evented_list import EventedList
from napari.utils.events.containers._nested_list import NestableEventedList
from napari.utils.events.containers._selectable_list import (
    SelectableEventedList,
    SelectableNestableEventedList,
)
from napari.utils.events.containers._selection import Selectable, Selection
from napari.utils.events.containers._set import EventedSet
from napari.utils.events.containers._typed import TypedMutableSequence

__all__ = [
    'EventedDict',
    'EventedList',
    'EventedSet',
    'NestableEventedList',
    'Selectable',
    'SelectableEventedList',
    'SelectableNestableEventedList',
    'Selection',
    'TypedMutableMapping',
    'TypedMutableSequence',
]



import pytest
from npe2 import DynamicPlugin
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QLabel, QPushButton

from napari._qt.widgets.qt_extension2reader import Extension2ReaderTable
from napari.settings import get_settings


@pytest.fixture
def extension2reader_widget(qtbot):
    def _extension2reader_widget(**kwargs):
        widget = Extension2ReaderTable(**kwargs)
        widget.show()
        qtbot.addWidget(widget)

        return widget

    return _extension2reader_widget


@pytest.fixture
def tif_reader(tmp_plugin: DynamicPlugin):
    tmp2 = tmp_plugin.spawn(name='tif_reader', register=True)

    @tmp2.contribute.reader(filename_patterns=['*.tif'])
    def _(path): ...

    return tmp2


@pytest.fixture
def npy_reader(tmp_plugin: DynamicPlugin):
    tmp2 = tmp_plugin.spawn(name='npy_reader', register=True)

    @tmp2.contribute.reader(filename_patterns=['*.npy'])
    def _(path): ...

    return tmp2


def test_extension2reader_defaults(
    extension2reader_widget,
):
    get_settings().plugins.extension2reader = {}
    widget = extension2reader_widget()

    assert widget._table.rowCount() == 1
    assert (
        widget._table.itemAt(0, 0).text() == 'No filename preferences found.'
    )


def test_extension2reader_with_settings(
    extension2reader_widget,
):
    get_settings().plugins.extension2reader = {'.test': 'test-plugin'}
    widget = extension2reader_widget()

    assert widget._table.rowCount() == 1
    assert widget._table.item(0, 0).text() == '.test'
    assert (
        widget._table.cellWidget(0, 1).findChild(QLabel).text()
        == 'test-plugin'
    )


def test_extension2reader_removal(extension2reader_widget, qtbot):
    get_settings().plugins.extension2reader = {
        '.test': 'test-plugin',
        '.abc': 'abc-plugin',
    }
    widget = extension2reader_widget()

    assert widget._table.rowCount() == 2

    btn_to_click = widget._table.cellWidget(0, 1).findChild(QPushButton)
    qtbot.mouseClick(btn_to_click, Qt.LeftButton)

    assert get_settings().plugins.extension2reader == {'.abc': 'abc-plugin'}
    assert widget._table.rowCount() == 1
    assert widget._table.item(0, 0).text() == '.abc'

    # remove remaining extension
    btn_to_click = widget._table.cellWidget(0, 1).findChild(QPushButton)
    qtbot.mouseClick(btn_to_click, Qt.LeftButton)
    assert not get_settings().plugins.extension2reader
    assert widget._table.rowCount() == 1
    assert 'No filename preferences found' in widget._table.item(0, 0).text()


def test_all_readers_in_dropdown(
    extension2reader_widget, tmp_plugin, tif_reader
):
    @tmp_plugin.contribute.reader(filename_patterns=['*'])
    def _(path): ...

    npe2_readers = {
        tmp_plugin.name: tmp_plugin.display_name,
        tif_reader.name: tif_reader.display_name,
    }

    widget = extension2reader_widget(npe2_readers=npe2_readers)
    all_dropdown_items = {
        widget._new_reader_dropdown.itemText(i)
        for i in range(widget._new_reader_dropdown.count())
    }
    assert all(i in all_dropdown_items for i in npe2_readers.values())


def test_directory_readers_not_in_dropdown(
    extension2reader_widget, tmp_plugin
):
    @tmp_plugin.contribute.reader(
        filename_patterns=[], accepts_directories=True
    )
    def f(path): ...

    widget = extension2reader_widget(
        npe2_readers={tmp_plugin.name: tmp_plugin.display_name},
        npe1_readers={},
    )
    all_dropdown_items = [
        widget._new_reader_dropdown.itemText(i)
        for i in range(widget._new_reader_dropdown.count())
    ]
    assert tmp_plugin.display_name not in all_dropdown_items


def test_filtering_readers(
    extension2reader_widget, builtins, tif_reader, npy_reader
):
    widget = extension2reader_widget(
        npe1_readers={builtins.display_name: builtins.display_name}
    )

    assert widget._new_reader_dropdown.count() == 3
    widget._filter_compatible_readers('*.npy')
    assert widget._new_reader_dropdown.count() == 2
    all_dropdown_items = [
        widget._new_reader_dropdown.itemText(i)
        for i in range(widget._new_reader_dropdown.count())
    ]
    assert (
        sorted([npy_reader.display_name, builtins.display_name])
        == all_dropdown_items
    )


@pytest.mark.parametrize('pattern', ['.', '', '/'])
def test_filtering_readers_problematic_patterns(
    extension2reader_widget, builtins, tif_reader, npy_reader, pattern
):
    widget = extension2reader_widget(
        npe1_readers={builtins.display_name: builtins.display_name}
    )
    widget._filter_compatible_readers(pattern)
    assert widget._new_reader_dropdown.count() == 1
    assert widget._new_reader_dropdown.itemText(0) == 'None available'


def test_filtering_readers_complex_pattern(
    extension2reader_widget, npy_reader, tif_reader
):
    @tif_reader.contribute.reader(
        filename_patterns=['my-specific-folder/*.tif']
    )
    def f(path): ...

    widget = extension2reader_widget(npe1_readers={})

    assert widget._new_reader_dropdown.count() == 2
    widget._filter_compatible_readers('my-specific-folder/my-file.tif')
    assert widget._new_reader_dropdown.count() == 1
    all_dropdown_items = [
        widget._new_reader_dropdown.itemText(i)
        for i in range(widget._new_reader_dropdown.count())
    ]
    assert sorted([tif_reader.name]) == all_dropdown_items


def test_adding_new_preference(
    extension2reader_widget, tif_reader, npy_reader
):
    widget = extension2reader_widget(npe1_readers={})
    widget._fn_pattern_edit.setText('*.tif')
    # will be filtered and tif-reader will be only item
    widget._new_reader_dropdown.setCurrentIndex(0)

    get_settings().plugins.extension2reader = {}
    widget._save_new_preference(True)
    settings = get_settings().plugins.extension2reader
    assert '*.tif' in settings
    assert settings['*.tif'] == tif_reader.name
    assert (
        widget._table.item(widget._table.rowCount() - 1, 0).text() == '*.tif'
    )
    plugin_label = widget._table.cellWidget(
        widget._table.rowCount() - 1, 1
    ).findChild(QLabel)
    assert plugin_label.text() == tif_reader.display_name


def test_adding_new_preference_no_asterisk(
    extension2reader_widget, tif_reader, npy_reader
):
    widget = extension2reader_widget(npe1_readers={})
    widget._fn_pattern_edit.setText('.tif')
    # will be filtered and tif-reader will be only item
    widget._new_reader_dropdown.setCurrentIndex(0)

    get_settings().plugins.extension2reader = {}
    widget._save_new_preference(True)
    settings = get_settings().plugins.extension2reader
    assert '*.tif' in settings
    assert settings['*.tif'] == tif_reader.name


def test_editing_preference(extension2reader_widget, tif_reader):
    tiff2 = tif_reader.spawn(register=True)

    @tiff2.contribute.reader(filename_patterns=['*.tif'])
    def ff(path): ...

    get_settings().plugins.extension2reader = {'*.tif': tif_reader.name}

    widget = extension2reader_widget()
    widget._fn_pattern_edit.setText('*.tif')
    # set to tiff2
    widget._new_reader_dropdown.setCurrentText(tiff2.display_name)
    original_row_count = widget._table.rowCount()
    widget._save_new_preference(True)
    settings = get_settings().plugins.extension2reader
    assert '*.tif' in settings
    assert settings['*.tif'] == tiff2.name
    assert widget._table.rowCount() == original_row_count
    plugin_label = widget._table.cellWidget(
        original_row_count - 1, 1
    ).findChild(QLabel)
    assert plugin_label.text() == tiff2.name



from napari._pydantic_compat import Field
from napari.utils.color import ColorValue
from napari.utils.events import EventedModel

_DEFAULT_COLOR = ColorValue('black')


class SurfaceWireframe(EventedModel):
    """
    Wireframe representation of the edges of a surface mesh.

    Attributes
    ----------
    visible : bool
        Whether the wireframe is displayed.
    color : ColorValue
        The color of the wireframe lines.
        See ``ColorValue.validate`` for supported values.
    width : float
        The width of the wireframe lines.
    """

    visible: bool = False
    color: ColorValue = Field(default_factory=lambda: _DEFAULT_COLOR)
    width: float = 1



import numpy as np

from napari._qt.layer_controls.qt_shapes_controls import QtShapesControls
from napari.layers import Shapes
from napari.utils.colormaps.standardize_color import transform_color

_SHAPES = np.random.random((10, 4, 2))


def test_shape_controls_face_color(qtbot):
    """Check updating of face color updates QtShapesControls."""
    layer = Shapes(_SHAPES)
    qtctrl = QtShapesControls(layer)
    qtbot.addWidget(qtctrl)
    target_color = transform_color(layer.current_face_color)[0]
    np.testing.assert_almost_equal(qtctrl.faceColorEdit.color, target_color)

    # Update current face color
    layer.current_face_color = 'red'
    target_color = transform_color(layer.current_face_color)[0]
    np.testing.assert_almost_equal(qtctrl.faceColorEdit.color, target_color)


def test_shape_controls_edge_color(qtbot):
    """Check updating of edge color updates QtShapesControls."""
    layer = Shapes(_SHAPES)
    qtctrl = QtShapesControls(layer)
    qtbot.addWidget(qtctrl)
    target_color = transform_color(layer.current_edge_color)[0]
    np.testing.assert_almost_equal(qtctrl.edgeColorEdit.color, target_color)

    # Update current edge color
    layer.current_edge_color = 'red'
    target_color = transform_color(layer.current_edge_color)[0]
    np.testing.assert_almost_equal(qtctrl.edgeColorEdit.color, target_color)


def test_text_visible_checkbox(qtbot):
    layer = Shapes(_SHAPES)
    qtctrl = QtShapesControls(layer)
    qtbot.addWidget(qtctrl)
    qtctrl.textDispCheckBox.setChecked(True)
    assert layer.text.visible
    qtctrl.textDispCheckBox.setChecked(False)
    assert not layer.text.visible
    qtctrl.textDispCheckBox.setChecked(True)
    assert layer.text.visible



import numpy as np

from napari._vispy.layers.base import VispyBaseLayer
from napari._vispy.visuals.vectors import VectorsVisual
from napari.layers.utils.layer_utils import segment_normal


class VispyVectorsLayer(VispyBaseLayer):
    def __init__(self, layer) -> None:
        node = VectorsVisual()
        super().__init__(layer, node)

        self.layer.events.edge_color.connect(self._on_data_change)

        self.reset()
        self._on_data_change()

    def _on_data_change(self):
        # Make meshes
        vertices, faces = generate_vector_meshes(
            self.layer._view_data,
            self.layer.edge_width,
            self.layer.length,
            self.layer.vector_style,
        )
        face_color = self.layer._view_face_color
        ndisplay = self.layer._slice_input.ndisplay
        ndim = self.layer.ndim

        if len(vertices) == 0 or len(faces) == 0:
            vertices = np.zeros((3, ndisplay))
            faces = np.array([[0, 1, 2]])
            face_color = np.array([[0, 0, 0, 0]])
        else:
            vertices = vertices[:, ::-1]

        if ndisplay == 3 and ndim == 2:
            vertices = np.pad(vertices, ((0, 0), (0, 1)), mode='constant')

        self.node.set_data(
            vertices=vertices,
            faces=faces,
            face_colors=face_color,
        )

        self.node.update()
        # Call to update order of translation values with new dims:
        self._on_matrix_change()


def generate_vector_meshes(vectors, width, length, vector_style):
    """Generates list of mesh vertices and triangles from a list of vectors

    Parameters
    ----------
    vectors : (N, 2, D) array
        A list of N vectors with start point and projections of the vector
        in D dimensions, where D is 2 or 3.
    width : float
        width of the vectors' bases
    length : float
        length multiplier of the line to be drawn
    vector_style : VectorStyle
        display style of the vectors

    Returns
    -------
    vertices : (aN, 2) array for 2D and (2aN, 2) array for 3D, with a=4, 3, or 7 for vector_style='line', 'triangle', or 'arrow' respectively
        Vertices of all triangles
    triangles : (bN, 3) array for 2D or (2bN, 3) array for 3D, with b=2, 1, or 3 for vector_style='line', 'triangle', or 'arrow' respectively
        Vertex indices that form the mesh triangles
    """
    ndim = vectors.shape[2]
    if ndim == 2:
        vertices, triangles = generate_vector_meshes_2D(
            vectors, width, length, vector_style
        )
    else:
        v_a, t_a = generate_vector_meshes_2D(
            vectors, width, length, vector_style, p=(0, 0, 1)
        )
        v_b, t_b = generate_vector_meshes_2D(
            vectors, width, length, vector_style, p=(1, 0, 0)
        )
        vertices = np.concatenate([v_a, v_b], axis=0)
        triangles = np.concatenate([t_a, len(v_a) + t_b], axis=0)

    return vertices, triangles


def generate_vector_meshes_2D(
    vectors, width, length, vector_style, p=(0, 0, 1)
):
    """Generates list of mesh vertices and triangles from a list of vectors

    Parameters
    ----------
    vectors : (N, 2, D) array
        A list of N vectors with start point and projections of the vector
        in D dimensions, where D is 2 or 3.
    width : float
        width of the vectors' bases
    length : float
        length multiplier of the line to be drawn
    vector_style : VectorStyle
        display style of the vectors
    p : 3-tuple, optional
        orthogonal vector for segment calculation in 3D.

    Returns
    -------
    vertices : (aN, 2) array for 2D, with a=4, 3, or 7 for vector_style='line', 'triangle', or 'arrow' respectively
        Vertices of all triangles
    triangles : (bN, 3) array for 2D, with b=2, 1, or 3 for vector_style='line', 'triangle', or 'arrow' respectively
        Vertex indices that form the mesh triangles
    """

    if vector_style == 'line':
        vertices, triangles = generate_meshes_line_2D(
            vectors, width, length, p
        )

    elif vector_style == 'triangle':
        vertices, triangles = generate_meshes_triangle_2D(
            vectors, width, length, p
        )

    elif vector_style == 'arrow':
        vertices, triangles = generate_meshes_arrow_2D(
            vectors, width, length, p
        )

    return vertices, triangles


def generate_meshes_line_2D(vectors, width, length, p):
    """Generates list of mesh vertices and triangles from a list of vectors.

    Vectors are composed of 4 vertices and 2 triangles.
    Vertices are generated according to the following scheme::

        1---x---0
        | .     |
        |   .   |
        |     . |
        3---v---2

    Where x marks the start point of the vector, and v its end point.

    In the case of k 2D vectors, the output 'triangles' is:
    [
        [0,1,2],                # vector 0,   triangle i=0
        [1,2,3],                # vector 0,   triangle i=1
        [4,5,6],                # vector 1,   triangle i=2
        [5,6,7],                # vector 1,   triangle i=3

        ...,

        [2i, 2i + 1, 2i + 2],   # vector k-1, triangle i=2k-2 (i%2=0)
        [2i - 1, 2i, 2i + 1]    # vector k-1, triangle i=2k-1 (i%2=1)
    ]

    Parameters
    ----------
    vectors : (N, 2, D) array
        A list of N vectors with start point and projections of the vector
        in D dimensions, where D is 2 or 3.
    width : float
        width of the vectors' bases
    length : float
        length multiplier of the line to be drawn
    p : 3-tuple
        orthogonal vector for segment calculation in 3D.

    Returns
    -------
    vertices : (4N, D) array
        Vertices of all triangles
    triangles : (2N, 3) array
        Vertex indices that form the mesh triangles
    """
    nvectors, _, ndim = vectors.shape

    vectors_starts = vectors[:, 0]
    vectors_ends = vectors_starts + length * vectors[:, 1]

    vertices = np.zeros((4 * nvectors, ndim))
    offsets = segment_normal(vectors_starts, vectors_ends, p=p)
    offsets = np.repeat(offsets, 4, axis=0)

    signs = np.ones((len(offsets), ndim))
    signs[::2] = -1
    offsets = offsets * signs

    vertices[::4] = vectors_starts
    vertices[1::4] = vectors_starts
    vertices[2::4] = vectors_ends
    vertices[3::4] = vectors_ends

    vertices = vertices + width * offsets / 2

    # Generate triangles in two steps:
    # 1. Repeat the vertices pattern
    # [[0,1,2],
    #  [1,2,3]]
    # as described in the docstring
    vertices_pattern = np.tile([[0, 1, 2], [1, 2, 3]], (nvectors, 1))
    # 2. Add an offset to differentiate between vectors
    triangles = (
        vertices_pattern + np.repeat(4 * np.arange(nvectors), 2)[:, np.newaxis]
    )

    triangles = triangles.astype(np.uint32)

    return vertices, triangles


def generate_meshes_triangle_2D(vectors, width, length, p):
    """Generate meshes forming 2D isosceles triangles to represent input vectors.

    Vectors are composed of 3 vertices and 1 triangles.
    Vertices are generated according to the following scheme::

        1---x---0
         .     .
          .   .
           . .
            2


    Where x marks the start point of the vector, and the vertex 2 its end
    point.

    In the case of k 2D vectors, the output 'triangles' is:
    [
        [0,1,2],                # vector 0,   triangle i=0
        [3,4,5],                # vector 1,   triangle i=1

        ...,

        [3i, 3i + 1, 3i + 2]    # vector k-1, triangle i=k-1
    ]

    Parameters
    ----------
    vectors : (N, 2, D) array
        A list of N vectors with start point and projections of the vector
        in D dimensions, where D is 2 or 3.
    width : float
        width of the vectors' bases
    length : float
        length multiplier of the line to be drawn
    p : 3-tuple
        orthogonal vector for segment calculation in 3D.

    Returns
    -------
    vertices : (3N, D) array
        Vertices of all triangles
    triangles : (N, 3) array
        Vertex indices that form the mesh triangles
    """
    nvectors, _, ndim = vectors.shape

    vectors_starts = vectors[:, 0]
    vectors_ends = vectors_starts + length * vectors[:, 1]

    vertices = np.zeros((3 * nvectors, ndim))
    offsets = segment_normal(vectors_starts, vectors_ends, p=p)
    offsets = np.repeat(offsets, 3, axis=0)

    signs = np.ones((len(offsets), ndim))
    signs[::3] = -1
    multipliers = np.ones((len(offsets), ndim))
    multipliers[2::3] = 0

    # here 'multipliers' is used to prevent vertex 2 from being offset
    offsets = offsets * signs * multipliers

    vertices[::3] = vectors_starts
    vertices[1::3] = vectors_starts
    vertices[2::3] = vectors_ends

    vertices = vertices + width * offsets / 2

    # faster than using the formula in the docstring
    triangles = np.arange(3 * nvectors, dtype=np.uint32).reshape((-1, 3))

    return vertices, triangles


def generate_meshes_arrow_2D(vectors, width, length, p):
    """Generate mesh forming 2D arrows given input vectors.

    Vectors are composed of 7 vertices and 3 triangles.
    Vertices are generated according to the following scheme::

            1---x---0
            | .     |
            |   .   |
            |     . |
        5---3-------2---4
           .         .
              .   .
                6

    Where x marks the start point of the vector, and the vertex 6 its end
    point.

    In the case of k 2D vectors, the output 'triangles' is:
    [
        [0,1,2],                # vector 0,   triangle i=0
        [1,2,3],                # vector 0,   triangle i=1
        [4,5,6],                # vector 0,   triangle i=2
        [7,8,9],                # vector 1,   triangle i=3
        [8,9,10],               # vector 1,   triangle i=4
        [11,12,13],             # vector 1,   triangle i=5

        ...,

        [7i/3,           7i/3 + 1,       7i/3 + 2],
            # vector k-1, triangle i=3k-3 (i%3=0)
        [7(i - 1)/3 + 1, 7(i - 1)/3 + 2, 7(i - 1)/3 + 3],
            # vector k-1, triangle i=3k-2 (i%3=1)
        [7(i - 2)/3 + 4, 7(i - 2)/3 + 5, 7(i - 2)/3 + 6]
            # vector k-1, triangle i=3k-1 (i%3=2)
    ]

    Parameters
    ----------
    vectors : (N, 2, D) array
        A list of N vectors with start point and projections of the vector
        in D dimensions, where D is 2 or 3.
    width : float
        width of the vectors' bases
    length : float
        length multiplier of the line to be drawn
    p : 3-tuple
        orthogonal vector for segment calculation in 3D.

    Returns
    -------
    vertices : (7N, D) array
        Vertices of all triangles
    triangles : (3N, 3) array
        Vertex indices that form the mesh triangles
    """
    nvectors, _, ndim = vectors.shape

    vectors_starts = vectors[:, 0]

    # Will be used to generate the vertices 2,3,4 and 5.
    # Right now the head of the arrow is put at 75% of the length
    # of the vector.
    vectors_intermediates = vectors_starts + 0.75 * length * vectors[:, 1]

    vectors_ends = vectors_starts + length * vectors[:, 1]

    vertices = np.zeros((7 * nvectors, ndim))
    offsets = segment_normal(vectors_starts, vectors_ends, p=p)
    offsets = np.repeat(offsets, 7, axis=0)

    signs = np.ones((len(offsets), ndim))
    signs[::2] = -1
    multipliers = np.ones((len(offsets), ndim))
    multipliers[4::7] = 2
    multipliers[5::7] = 2
    multipliers[6::7] = 0

    # here 'multipliers' is used to prevent vertex 6 from being offset,
    # and to offset vertices 4 and 5 twice as much as vertices 2 and 3
    offsets = offsets * signs * multipliers

    vertices[::7] = vectors_starts
    vertices[1::7] = vectors_starts
    vertices[2::7] = vectors_intermediates
    vertices[3::7] = vectors_intermediates
    vertices[4::7] = vectors_intermediates
    vertices[5::7] = vectors_intermediates
    vertices[6::7] = vectors_ends

    vertices = vertices + width * offsets / 2

    # Generate triangles in two steps:
    # 1. Repeat the vertices pattern
    # [[0,1,2],
    #  [1,2,3]
    #  [4,5,6]]
    # as described in the docstring
    vertices_pattern = np.tile(
        [[0, 1, 2], [1, 2, 3], [4, 5, 6]], (nvectors, 1)
    )
    # 2. Add an offset to differentiate between vectors
    triangles = (
        vertices_pattern + np.repeat(7 * np.arange(nvectors), 3)[:, np.newaxis]
    )

    triangles = triangles.astype(np.uint32)

    return vertices, triangles



import numpy as np
import pytest

from napari.layers.utils.interactivity_utils import (
    drag_data_to_projected_distance,
)


@pytest.mark.parametrize(
    (
        'start_position',
        'end_position',
        'view_direction',
        'vector',
        'expected_value',
    ),
    [
        # drag vector parallel to view direction
        # projected onto perpendicular vector
        ([0, 0, 0], [0, 0, 1], [0, 0, 1], [1, 0, 0], 0),
        # same as above, projection onto multiple perpendicular vectors
        # should produce multiple results
        ([0, 0, 0], [0, 0, 1], [0, 0, 1], [[1, 0, 0], [0, 1, 0]], [0, 0]),
        # drag vector perpendicular to view direction
        # projected onto itself
        ([0, 0, 0], [0, 1, 0], [0, 0, 1], [0, 1, 0], 1),
    ],
)
def test_drag_data_to_projected_distance(
    start_position, end_position, view_direction, vector, expected_value
):
    result = drag_data_to_projected_distance(
        start_position, end_position, view_direction, vector
    )
    assert np.allclose(result, expected_value)



import numpy as np
import pytest


def test_big_2D_image(make_napari_viewer):
    """Test big 2D image with axis exceeding max texture size."""
    viewer = make_napari_viewer()

    shape = (20_000, 10)
    data = np.random.random(shape)
    layer = viewer.add_image(data, multiscale=False)
    visual = viewer.window._qt_viewer.canvas.layer_to_visual[layer]
    assert visual.node is not None
    if visual.MAX_TEXTURE_SIZE_2D is not None:
        s = np.ceil(np.divide(shape, visual.MAX_TEXTURE_SIZE_2D)).astype(int)
        np.testing.assert_array_equal(layer._transforms['tile2data'].scale, s)


def test_big_3D_image(make_napari_viewer):
    """Test big 3D image with axis exceeding max texture size."""
    viewer = make_napari_viewer(ndisplay=3)

    shape = (5, 10, 3_000)
    data = np.random.random(shape)
    layer = viewer.add_image(data, multiscale=False)
    visual = viewer.window._qt_viewer.canvas.layer_to_visual[layer]
    assert visual.node is not None
    if visual.MAX_TEXTURE_SIZE_3D is not None:
        s = np.ceil(np.divide(shape, visual.MAX_TEXTURE_SIZE_3D)).astype(int)
        np.testing.assert_array_equal(layer._transforms['tile2data'].scale, s)


@pytest.mark.parametrize(
    'shape',
    [(2, 4), (256, 4048), (4, 20_000), (20_000, 4)],
)
def test_downsample_value(make_napari_viewer, shape):
    """Test getting correct value for downsampled data."""
    viewer = make_napari_viewer()

    data = np.zeros(shape)
    data[shape[0] // 2 :, shape[1] // 2 :] = 1
    layer = viewer.add_image(data, multiscale=False)

    test_points = [
        (int(shape[0] * 0.25), int(shape[1] * 0.25)),
        (int(shape[0] * 0.75), int(shape[1] * 0.25)),
        (int(shape[0] * 0.25), int(shape[1] * 0.75)),
        (int(shape[0] * 0.75), int(shape[1] * 0.75)),
    ]
    expected_values = [0.0, 0.0, 0.0, 1.0]

    for test_point, expected_value in zip(test_points, expected_values):
        viewer.cursor.position = test_point
        assert (
            layer.get_value(viewer.cursor.position, world=True)
            == expected_value
        )



import numpy as np

from napari._qt.layer_controls.qt_surface_controls import QtSurfaceControls
from napari.layers import Surface
from napari.layers.surface._surface_constants import SHADING_TRANSLATION

data = np.array([[0, 0], [0, 20], [10, 0], [10, 10]])
faces = np.array([[0, 1, 2], [1, 2, 3]])
values = np.linspace(0, 1, len(data))
_SURFACE = (data, faces, values)


def test_shading_combobox(qtbot):
    layer = Surface(_SURFACE)
    qtctrl = QtSurfaceControls(layer)
    qtbot.addWidget(qtctrl)
    assert qtctrl.shadingComboBox.currentText() == layer.shading

    for display, shading in SHADING_TRANSLATION.items():
        qtctrl.shadingComboBox.setCurrentText(display)
        assert layer.shading == shading

    for display, shading in SHADING_TRANSLATION.items():
        layer.shading = shading
        assert qtctrl.shadingComboBox.currentText() == display



from __future__ import annotations

from functools import lru_cache, partial
from itertools import chain
from typing import TYPE_CHECKING

from napari._qt._qapp_model.injection._qprocessors import QPROCESSORS
from napari._qt._qapp_model.injection._qproviders import QPROVIDERS

if TYPE_CHECKING:
    from app_model.types import Context

# Submodules should be able to import from most modules, so to
# avoid circular imports, don't import submodules at the top level here,
# import them inside the init_qactions function.


@lru_cache  # only call once
def init_qactions() -> None:
    """Initialize all Qt-based Actions with app-model

    This function will be called in _QtMainWindow.__init__().  It should only
    be called once (hence the lru_cache decorator).

    It is responsible for:
    - injecting Qt-specific names into the application injection_store namespace
      (this is what allows functions to be declared with annotations like
      `def foo(window: Window)` or `def foo(qt_viewer: QtViewer)`)
    - registering provider functions for the names added to the namespace
    - registering Qt-dependent actions with app-model (i.e. Q_*_ACTIONS actions).
    """
    from napari._app_model import get_app_model
    from napari._qt._qapp_model.qactions._debug import (
        DEBUG_SUBMENUS,
        Q_DEBUG_ACTIONS,
    )
    from napari._qt._qapp_model.qactions._file import (
        FILE_SUBMENUS,
        Q_FILE_ACTIONS,
    )
    from napari._qt._qapp_model.qactions._help import Q_HELP_ACTIONS
    from napari._qt._qapp_model.qactions._layerlist_context import (
        Q_LAYERLIST_CONTEXT_ACTIONS,
    )
    from napari._qt._qapp_model.qactions._layers_actions import (
        LAYERS_ACTIONS,
        LAYERS_SUBMENUS,
    )
    from napari._qt._qapp_model.qactions._plugins import Q_PLUGINS_ACTIONS
    from napari._qt._qapp_model.qactions._view import (
        Q_VIEW_ACTIONS,
        VIEW_SUBMENUS,
    )
    from napari._qt._qapp_model.qactions._window import Q_WINDOW_ACTIONS
    from napari._qt.qt_main_window import Window
    from napari._qt.qt_viewer import QtViewer

    # update the namespace with the Qt-specific types/providers/processors
    app = get_app_model()
    store = app.injection_store
    store.namespace = {
        **store.namespace,
        'Window': Window,
        'QtViewer': QtViewer,
    }

    # Qt-specific providers/processors
    app.injection_store.register(
        processors=QPROCESSORS,
        providers=QPROVIDERS,
    )

    # register menubar actions
    app.register_actions(
        chain(
            Q_DEBUG_ACTIONS,
            Q_FILE_ACTIONS,
            Q_HELP_ACTIONS,
            Q_PLUGINS_ACTIONS,
            Q_VIEW_ACTIONS,
            LAYERS_ACTIONS,
            Q_LAYERLIST_CONTEXT_ACTIONS,
            Q_WINDOW_ACTIONS,
        )
    )

    # register menubar submenus
    app.menus.append_menu_items(
        chain(FILE_SUBMENUS, VIEW_SUBMENUS, DEBUG_SUBMENUS, LAYERS_SUBMENUS)
    )


def add_dummy_actions(context: Context) -> None:
    """Register dummy 'Empty' actions for all contributable menus.

    Each action is registered with its own `when` condition, that
    ensures the action is not visible once the menu is populated.
    The context key used in the `when` condition is also added to
    the given `context` and assigned to a partial function that
    returns True if the menu is empty, and otherwise False.


    Parameters
    ----------
    context : Context
        context to store functional keys used in `when` conditions
    """
    from napari._app_model import get_app_model
    from napari._app_model.constants._menus import MenuId
    from napari._app_model.utils import get_dummy_action, is_empty_menu

    app = get_app_model()

    actions = []
    for menu_id in MenuId.contributables():
        dummmy_action, context_key = get_dummy_action(menu_id)
        if dummmy_action.id not in app.commands:
            actions.append(dummmy_action)
        # NOTE: even if action is already registered, the `context` instance
        # may be new e.g. when closing and relaunching a viewer
        # in a notebook. Context key should be assigned regardless
        context[context_key] = partial(is_empty_menu, menu_id)
    app.register_actions(actions)



import warnings
from unittest.mock import patch

import pytest
from napari_plugin_engine import napari_hook_implementation

from napari import Viewer
from napari._qt import Window
from napari._tests.utils import skip_on_win_ci
from napari.settings import get_settings
from napari.utils.theme import Theme, get_theme


@skip_on_win_ci
@patch.object(Window, '_remove_theme')
@patch.object(Window, '_add_theme')
def test_provide_theme_hook_registered_correctly(
    mock_add_theme,
    mock_remove_theme,
    make_napari_viewer,
    napari_plugin_manager,
):
    # make a viewer with a plugin & theme registered
    viewer = make_napari_viewer_with_plugin_theme(
        make_napari_viewer,
        napari_plugin_manager,
        theme_type='dark',
        name='dark-test-2',
    )

    # set the viewer theme to the plugin theme
    viewer.theme = 'dark-test-2'

    # triggered when theme was added
    mock_add_theme.assert_called()
    mock_remove_theme.assert_not_called()

    # now, lets unregister the theme
    # We didn't set the setting, so ensure that no warning
    with warnings.catch_warnings():
        warnings.simplefilter('error')
        napari_plugin_manager.unregister('TestPlugin')
    mock_remove_theme.assert_called()


@patch.object(Window, '_remove_theme')
@patch.object(Window, '_add_theme')
def test_plugin_provide_theme_hook_set_settings_correctly(
    mock_add_theme,
    mock_remove_theme,
    make_napari_viewer,
    napari_plugin_manager,
):
    # make a viewer with a plugin & theme registered
    make_napari_viewer_with_plugin_theme(
        make_napari_viewer,
        napari_plugin_manager,
        theme_type='dark',
        name='dark-test-2',
    )
    # set the plugin theme as a setting
    get_settings().appearance.theme = 'dark-test-2'

    # triggered when theme was added
    mock_add_theme.assert_called()
    mock_remove_theme.assert_not_called()

    # now, lets unregister the theme
    # We *did* set the setting, so there should be a warning
    with pytest.warns(UserWarning, match='The current theme '):
        napari_plugin_manager.unregister('TestPlugin')
    mock_remove_theme.assert_called()


def make_napari_viewer_with_plugin_theme(
    make_napari_viewer, napari_plugin_manager, *, theme_type: str, name: str
) -> Viewer:
    theme = get_theme(theme_type).to_rgb_dict()
    theme['name'] = name

    class TestPlugin:
        @napari_hook_implementation
        def napari_experimental_provide_theme():
            return {name: theme}

    # create instance of viewer to make sure
    # registration and unregistration methods are called
    viewer = make_napari_viewer()

    # register theme
    napari_plugin_manager.register(TestPlugin)
    reg = napari_plugin_manager._theme_data['TestPlugin']
    assert isinstance(reg[name], Theme)

    return viewer



import gc
from unittest.mock import Mock

from napari.utils.events.event import Event, EventEmitter
from napari.utils.events.event_utils import (
    connect_no_arg,
    connect_setattr,
    connect_setattr_value,
)


def test_connect_no_arg():
    mock = Mock(['meth'])
    emiter = EventEmitter()
    connect_no_arg(emiter, mock, 'meth')
    emiter(type_name='a', value=1)
    mock.meth.assert_called_once_with()
    assert len(emiter.callbacks) == 1
    del mock
    gc.collect()
    assert len(emiter.callbacks) == 1
    emiter(type_name='a', value=1)
    assert len(emiter.callbacks) == 0


def test_connect_setattr_value():
    mock = Mock()
    emiter = EventEmitter()
    connect_setattr_value(emiter, mock, 'meth')
    emiter(type_name='a', value=1)
    assert mock.meth == 1
    assert len(emiter.callbacks) == 1
    del mock
    gc.collect()
    assert len(emiter.callbacks) == 1
    emiter(type_name='a', value=1)
    assert len(emiter.callbacks) == 0


def test_connect_setattr():
    mock = Mock()
    emiter = EventEmitter()
    connect_setattr(emiter, mock, 'meth')
    emiter(type_name='a', value=1)
    assert isinstance(mock.meth, Event)
    assert mock.meth.value == 1
    assert len(emiter.callbacks) == 1
    del mock
    gc.collect()
    assert len(emiter.callbacks) == 1
    emiter(type_name='a', value=1)
    assert len(emiter.callbacks) == 0



from __future__ import annotations

import logging
import sys
import traceback
import warnings
import weakref
from collections.abc import Sequence
from pathlib import Path
from types import FrameType
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Union,
)
from weakref import WeakSet, ref

import numpy as np
from qtpy.QtCore import QCoreApplication, QObject, Qt, QUrl
from qtpy.QtGui import QGuiApplication
from qtpy.QtWidgets import QFileDialog, QSplitter, QVBoxLayout, QWidget
from superqt import ensure_main_thread

from napari._qt.containers import QtLayerList
from napari._qt.dialogs.qt_reader_dialog import handle_gui_reading
from napari._qt.dialogs.screenshot_dialog import ScreenshotDialog
from napari._qt.perf.qt_performance import QtPerformance
from napari._qt.utils import QImg2array
from napari._qt.widgets.qt_dims import QtDims
from napari._qt.widgets.qt_viewer_buttons import (
    QtLayerButtons,
    QtViewerButtons,
)
from napari._qt.widgets.qt_viewer_dock_widget import QtViewerDockWidget
from napari._qt.widgets.qt_welcome import QtWidgetOverlay
from napari.components.camera import Camera
from napari.components.layerlist import LayerList
from napari.errors import MultipleReaderError, ReaderPluginError
from napari.layers.base.base import Layer
from napari.plugins import _npe2
from napari.settings import get_settings
from napari.settings._application import DaskSettings
from napari.utils import config, perf, resize_dask_cache
from napari.utils.action_manager import action_manager
from napari.utils.history import (
    get_open_history,
    get_save_history,
    update_open_history,
    update_save_history,
)
from napari.utils.io import imsave
from napari.utils.key_bindings import KeymapHandler
from napari.utils.misc import in_ipython, in_jupyter
from napari.utils.naming import CallerFrame
from napari.utils.notifications import show_info
from napari.utils.translations import trans
from napari_builtins.io import imsave_extensions

from napari._vispy import VispyCanvas, create_vispy_layer  # isort:skip

if TYPE_CHECKING:
    from napari_console import QtConsole
    from npe2.manifest.contributions import WriterContribution

    from napari._qt.layer_controls import QtLayerControlsContainer
    from napari.components import ViewerModel
    from napari.utils.events import Event


def _npe2_decode_selected_filter(
    ext_str: str, selected_filter: str, writers: Sequence[WriterContribution]
) -> Optional[WriterContribution]:
    """Determine the writer that should be invoked to save data.

    When npe2 can be imported, resolves a selected file extension
    string into a specific writer. Otherwise, returns None.
    """
    # When npe2 is not present, `writers` is expected to be an empty list,
    # `[]`. This function will return None.

    for entry, writer in zip(
        ext_str.split(';;'),
        writers,
    ):
        if entry.startswith(selected_filter):
            return writer
    return None


def _extension_string_for_layers(
    layers: Sequence[Layer],
) -> tuple[str, list[WriterContribution]]:
    """Return an extension string and the list of corresponding writers.

    The extension string is a ";;" delimeted string of entries. Each entry
    has a brief description of the file type and a list of extensions.

    The writers, when provided, are the npe2.manifest.io.WriterContribution
    objects. There is one writer per entry in the extension string. If npe2
    is not importable, the list of writers will be empty.
    """
    # try to use npe2
    ext_str, writers = _npe2.file_extensions_string_for_layers(layers)
    if ext_str:
        return ext_str, writers

    # fallback to old behavior

    if len(layers) == 1:
        selected_layer = layers[0]
        # single selected layer.
        if selected_layer._type_string == 'image':
            ext = imsave_extensions()

            ext_list = [f'*{val}' for val in ext]
            ext_str = ';;'.join(ext_list)

            ext_str = trans._(
                'All Files (*);; Image file types:;;{ext_str}',
                ext_str=ext_str,
            )

        elif selected_layer._type_string == 'points':
            ext_str = trans._('All Files (*);; *.csv;;')

        else:
            # layer other than image or points
            ext_str = trans._('All Files (*);;')

    else:
        # multiple layers.
        ext_str = trans._('All Files (*);;')
    return ext_str, []


class QtViewer(QSplitter):
    """Qt view for the napari Viewer model.

    Parameters
    ----------
    viewer : napari.components.ViewerModel
        Napari viewer containing the rendered scene, layers, and controls.
    show_welcome_screen : bool, optional
        Flag to show a welcome message when no layers are present in the
        canvas. Default is `False`.
    canvas_class : napari._vispy.canvas.VispyCanvas
        The VispyCanvas class providing the Vispy SceneCanvas. Users can also
        have a custom canvas here.

    Attributes
    ----------
    canvas : napari._vispy.canvas.VispyCanvas
        The VispyCanvas class providing the Vispy SceneCanvas. Users can also
        have a custom canvas here.
    dims : napari.qt_dims.QtDims
        Dimension sliders; Qt View for Dims model.
    show_welcome_screen : bool
        Boolean indicating whether to show the welcome screen.
    viewer : napari.components.ViewerModel
        Napari viewer containing the rendered scene, layers, and controls.
    _key_map_handler : napari.utils.key_bindings.KeymapHandler
        KeymapHandler handling the calling functionality when keys are pressed that have a callback function mapped
    _qt_poll : Optional[napari._qt.experimental.qt_poll.QtPoll]
        A QtPoll object required for the monitor.
    _remote_manager : napari.components.experimental.remote.RemoteManager
        A remote manager processing commands from remote clients and sending out messages when polled.
    _welcome_widget : napari._qt.widgets.qt_welcome.QtWidgetOverlay
        QtWidgetOverlay providing the stacked widgets for the welcome page.
    """

    _instances = WeakSet()

    def __init__(
        self,
        viewer: ViewerModel,
        show_welcome_screen: bool = False,
        canvas_class: type[VispyCanvas] = VispyCanvas,
    ) -> None:
        super().__init__()
        self._instances.add(self)
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)

        self._show_welcome_screen = show_welcome_screen

        QCoreApplication.setAttribute(
            Qt.AA_UseStyleSheetPropagationInWidgetStyles, True
        )

        self.viewer = viewer
        self.dims = QtDims(self.viewer.dims)
        self._controls = None
        self._layers = None
        self._layersButtons = None
        self._viewerButtons = None
        self._key_map_handler = KeymapHandler()
        self._key_map_handler.keymap_providers = [self.viewer]
        self._console_backlog = []
        self._console = None

        self._dockLayerList = None
        self._dockLayerControls = None
        self._dockConsole = None
        self._dockPerformance = None

        # This dictionary holds the corresponding vispy visual for each layer
        self.canvas = canvas_class(
            viewer=viewer,
            parent=self,
            key_map_handler=self._key_map_handler,
            size=self.viewer._canvas_size,
            autoswap=get_settings().experimental.autoswap_buffers,  # see #5734
        )

        # Stacked widget to provide a welcome page
        self._welcome_widget = QtWidgetOverlay(self, self.canvas.native)
        self._welcome_widget.set_welcome_visible(show_welcome_screen)
        self._welcome_widget.sig_dropped.connect(self.dropEvent)
        self._welcome_widget.leave.connect(self._leave_canvas)
        self._welcome_widget.enter.connect(self._enter_canvas)

        main_widget = QWidget()
        main_layout = QVBoxLayout()
        main_layout.setContentsMargins(0, 2, 0, 2)
        main_layout.addWidget(self._welcome_widget)
        main_layout.addWidget(self.dims)
        main_layout.setSpacing(0)
        main_widget.setLayout(main_layout)

        self.setOrientation(Qt.Orientation.Vertical)
        self.addWidget(main_widget)

        self.viewer._layer_slicer.events.ready.connect(self._on_slice_ready)

        self._on_active_change()
        self.viewer.layers.events.inserted.connect(self._update_welcome_screen)
        self.viewer.layers.events.removed.connect(self._update_welcome_screen)
        self.viewer.layers.selection.events.active.connect(
            self._on_active_change
        )

        self.viewer.layers.events.inserted.connect(self._on_add_layer_change)

        self.setAcceptDrops(True)

        # Create the experimental QtPool for the monitor.
        self._qt_poll = _create_qt_poll(self, self.viewer.camera)

        # Create the experimental RemoteManager for the monitor.
        self._remote_manager = _create_remote_manager(
            self.viewer.layers, self._qt_poll
        )

        # bind shortcuts stored in settings last.
        self._bind_shortcuts()

        settings = get_settings()
        self._update_dask_cache_settings(settings.application.dask)

        settings.application.dask.events.connect(
            self._update_dask_cache_settings
        )

        for layer in self.viewer.layers:
            self._add_layer(layer)

    @property
    def view(self):
        """
        Rectangular  vispy viewbox widget in which a subscene is rendered. Access directly within the QtViewer will
        become deprecated.
        """
        warnings.warn(
            trans._(
                'Access to QtViewer.view is deprecated since 0.5.0 and will be removed in the napari 0.6.0. Change to QtViewer.canvas.view instead.'
            ),
            FutureWarning,
            stacklevel=2,
        )
        return self.canvas.view

    @property
    def camera(self):
        """
        The Vispy camera class which contains both the 2d and 3d camera used to describe the perspective by which a
        scene is viewed and interacted with. Access directly within the QtViewer will become deprecated.
        """
        warnings.warn(
            trans._(
                'Access to QtViewer.camera will become deprecated in the 0.6.0. Change to QtViewer.canvas.camera instead.'
            ),
            FutureWarning,
            stacklevel=2,
        )
        return self.canvas.camera

    @property
    def chunk_receiver(self) -> None:
        warnings.warn(
            trans._(
                'QtViewer.chunk_receiver is deprecated in version 0.5 and will be removed in a later version. '
                'More generally the old approach to async loading was removed in version 0.5 so this value is always None. '
                'If you need to specifically use the old approach, continue to use the latest 0.4 release.'
            ),
            DeprecationWarning,
        )

    @staticmethod
    def _update_dask_cache_settings(
        dask_setting: Union[DaskSettings, Event] = None,
    ):
        """Update dask cache to match settings."""
        if not dask_setting:
            return
        if not isinstance(dask_setting, DaskSettings):
            dask_setting = get_settings().application.dask

        enabled = dask_setting.enabled
        size = dask_setting.cache
        resize_dask_cache(int(int(enabled) * size * 1e9))

    @property
    def controls(self) -> QtLayerControlsContainer:
        """Qt view for GUI controls."""
        if self._controls is None:
            # Avoid circular import.
            from napari._qt.layer_controls import QtLayerControlsContainer

            self._controls = QtLayerControlsContainer(self.viewer)
        return self._controls

    @property
    def layers(self) -> QtLayerList:
        """Qt view for LayerList controls."""
        if self._layers is None:
            self._layers = QtLayerList(self.viewer.layers)
        return self._layers

    @property
    def layerButtons(self) -> QtLayerButtons:
        """Button controls for napari layers."""
        if self._layersButtons is None:
            self._layersButtons = QtLayerButtons(self.viewer)
        return self._layersButtons

    @property
    def viewerButtons(self) -> QtViewerButtons:
        """Button controls for the napari viewer."""
        if self._viewerButtons is None:
            self._viewerButtons = QtViewerButtons(self.viewer)
        return self._viewerButtons

    @property
    def dockLayerList(self) -> QtViewerDockWidget:
        """QWidget wrapped in a QDockWidget with forwarded viewer events."""
        if self._dockLayerList is None:
            layerList = QWidget()
            layerList.setObjectName('layerList')
            layerListLayout = QVBoxLayout()
            layerListLayout.addWidget(self.layerButtons)
            layerListLayout.addWidget(self.layers)
            layerListLayout.addWidget(self.viewerButtons)
            layerListLayout.setContentsMargins(8, 4, 8, 6)
            layerList.setLayout(layerListLayout)
            self._dockLayerList = QtViewerDockWidget(
                self,
                layerList,
                name=trans._('layer list'),
                area='left',
                allowed_areas=['left', 'right'],
                object_name='layer list',
                close_btn=False,
            )
        return self._dockLayerList

    @property
    def dockLayerControls(self) -> QtViewerDockWidget:
        """QWidget wrapped in a QDockWidget with forwarded viewer events."""
        if self._dockLayerControls is None:
            self._dockLayerControls = QtViewerDockWidget(
                self,
                self.controls,
                name=trans._('layer controls'),
                area='left',
                allowed_areas=['left', 'right'],
                object_name='layer controls',
                close_btn=False,
            )
        return self._dockLayerControls

    @property
    def dockConsole(self) -> QtViewerDockWidget:
        """QWidget wrapped in a QDockWidget with forwarded viewer events."""
        if self._dockConsole is None:
            self._dockConsole = QtViewerDockWidget(
                self,
                QWidget(),
                name=trans._('console'),
                area='bottom',
                allowed_areas=['top', 'bottom'],
                object_name='console',
                close_btn=False,
            )
            self._dockConsole.setVisible(False)
            self._dockConsole.visibilityChanged.connect(self._ensure_connect)
        return self._dockConsole

    @property
    def dockPerformance(self) -> QtViewerDockWidget:
        if self._dockPerformance is None:
            self._dockPerformance = self._create_performance_dock_widget()
        return self._dockPerformance

    @property
    def layer_to_visual(self):
        """Mapping of Napari layer to Vispy layer. Added for backward compatibility"""
        return self.canvas.layer_to_visual

    def _leave_canvas(self):
        """disable status on canvas leave"""
        self.viewer.status = ''
        self.viewer.mouse_over_canvas = False

    def _enter_canvas(self):
        """enable status on canvas enter"""
        self.viewer.status = 'Ready'
        self.viewer.mouse_over_canvas = True

    def _ensure_connect(self):
        # lazy load console
        id(self.console)

    def _bind_shortcuts(self):
        """Bind shortcuts stored in SETTINGS to actions."""
        for action, shortcuts in get_settings().shortcuts.shortcuts.items():
            action_manager.unbind_shortcut(action)
            for shortcut in shortcuts:
                action_manager.bind_shortcut(action, shortcut)

    def _create_performance_dock_widget(self):
        """Create the dock widget that shows performance metrics."""
        if perf.perf_config is not None:
            return QtViewerDockWidget(
                self,
                QtPerformance(),
                name=trans._('performance'),
                area='bottom',
            )
        return None

    def _weakref_if_possible(self, obj):
        """Create a weakref to obj.

        Parameters
        ----------
        obj : object
            Cannot create weakrefs to many Python built-in datatypes such as
            list, dict, str.

            From https://docs.python.org/3/library/weakref.html: "Objects which
            support weak references include class instances, functions written
            in Python (but not in C), instance methods, sets, frozensets, some
            file objects, generators, type objects, sockets, arrays, deques,
            regular expression pattern objects, and code objects."

        Returns
        -------
        weakref or object
            Returns a weakref if possible.
        """
        try:
            newref = ref(obj)
        except TypeError:
            newref = obj
        return newref

    def _unwrap_if_weakref(self, value):
        """Return value or if that is weakref the object referenced by value.

        Parameters
        ----------
        value : object or weakref
            No-op for types other than weakref.

        Returns
        -------
        unwrapped: object or None
            Returns referenced object, or None if weakref is dead.
        """
        unwrapped = value() if isinstance(value, ref) else value
        return unwrapped

    def add_to_console_backlog(self, variables):
        """Save variables for pushing to console when it is instantiated.

        This function will create weakrefs when possible to avoid holding on to
        too much memory unnecessarily.

        Parameters
        ----------
        variables : dict, str or list/tuple of str
            The variables to inject into the console's namespace. If a dict, a
            simple update is done. If a str, the string is assumed to have
            variable names separated by spaces. A list/tuple of str can also
            be used to give the variable names. If just the variable names are
            give (list/tuple/str) then the variable values looked up in the
            callers frame.
        """
        if isinstance(variables, (str, list, tuple)):
            if isinstance(variables, str):
                vlist = variables.split()
            else:
                vlist = variables
            vdict = {}
            cf = sys._getframe(2)
            for name in vlist:
                try:
                    vdict[name] = eval(name, cf.f_globals, cf.f_locals)
                except NameError:
                    logging.warning(
                        'Could not get variable %s from %s',
                        name,
                        cf.f_code.co_name,
                    )
        elif isinstance(variables, dict):
            vdict = variables
        else:
            raise TypeError('variables must be a dict/str/list/tuple')
        # weakly reference values if possible
        new_dict = {k: self._weakref_if_possible(v) for k, v in vdict.items()}
        self.console_backlog.append(new_dict)

    @property
    def console_backlog(self):
        """List: items to push to console when instantiated."""
        return self._console_backlog

    def _get_console(self) -> Optional[QtConsole]:
        """Function to setup console.

        Returns
        -------
        console : QtConsole or None
            The napari console.

        Notes
        _____
        _get_console extracted to separate function to simplify testing.

        """
        try:
            import numpy as np

            # QtConsole imports debugpy that overwrites default breakpoint.
            # It makes problems with debugging if you do not know this.
            # So we do not want to overwrite it if it is already set.
            breakpoint_handler = sys.breakpointhook
            from napari_console import QtConsole

            sys.breakpointhook = breakpoint_handler

            import napari

            with warnings.catch_warnings():
                warnings.filterwarnings('ignore')
                console = QtConsole(self.viewer, style_sheet=self.styleSheet())
                console.push(
                    {'napari': napari, 'action_manager': action_manager}
                )
                with CallerFrame(_in_napari) as c:
                    if c.frame.f_globals.get('__name__', '') == '__main__':
                        console.push({'np': np})
                for i in self.console_backlog:
                    # recover weak refs
                    console.push(
                        {
                            k: self._unwrap_if_weakref(v)
                            for k, v in i.items()
                            if self._unwrap_if_weakref(v) is not None
                        }
                    )
                return console
        except ModuleNotFoundError:
            warnings.warn(
                trans._(
                    'napari-console not found. It can be installed with'
                    ' "pip install napari_console"'
                ),
                stacklevel=1,
            )
            return None
        except ImportError:
            traceback.print_exc()
            warnings.warn(
                trans._(
                    'error importing napari-console. See console for full error.'
                ),
                stacklevel=1,
            )
            return None

    @property
    def console(self):
        """QtConsole: iPython console terminal integrated into the napari GUI."""
        if self._console is None:
            self.console = self._get_console()
            self._console_backlog = []
        return self._console

    @console.setter
    def console(self, console):
        self._console = console
        if console is not None:
            self.dockConsole.setWidget(console)
            console.setParent(self.dockConsole)

    @ensure_main_thread
    def _on_slice_ready(self, event):
        """Callback connected to `viewer._layer_slicer.events.ready`.

        Provides updates after slicing using the slice response data.
        This only gets triggered on the async slicing path.
        """
        responses: dict[weakref.ReferenceType[Layer], Any] = event.value
        logging.debug('QtViewer._on_slice_ready: %s', responses)
        for weak_layer, response in responses.items():
            if layer := weak_layer():
                # Update the layer slice state to temporarily support behavior
                # that depends on it.
                layer._update_slice_response(response)
                # Update the layer's loaded state before everything else,
                # because they may rely on its updated value.
                layer._update_loaded_slice_id(response.request_id)
                # The rest of `Layer.refresh` after `set_view_slice`, where
                # `set_data` notifies the corresponding vispy layer of the new
                # slice.
                layer.events.set_data()
                layer._update_thumbnail()
                layer._set_highlight(force=True)

    def _on_active_change(self):
        """When active layer changes change keymap handler."""
        self._key_map_handler.keymap_providers = (
            [self.viewer]
            if self.viewer.layers.selection.active is None
            else [self.viewer.layers.selection.active, self.viewer]
        )

    def _on_add_layer_change(self, event):
        """When a layer is added, set its parent and order.

        Parameters
        ----------
        event : napari.utils.event.Event
            The napari event that triggered this method.
        """
        layer = event.value
        self._add_layer(layer)

    def _add_layer(self, layer):
        """When a layer is added, set its parent and order.

        Parameters
        ----------
        layer : napari.layers.Layer
            Layer to be added.
        """
        vispy_layer = create_vispy_layer(layer)

        # QtPoll is experimental.
        if self._qt_poll is not None:
            # QtPoll will call VipyBaseImage._on_poll() when the camera
            # moves or the timer goes off.
            self._qt_poll.events.poll.connect(vispy_layer._on_poll)

            # In the other direction, some visuals need to tell QtPoll to
            # start polling. When they receive new data they need to be
            # polled to load it, even if the camera is not moving.
            if vispy_layer.events is not None:
                vispy_layer.events.loaded.connect(self._qt_poll.wake_up)

        self.canvas.add_layer_visual_mapping(layer, vispy_layer)

    def _remove_invalid_chars(self, selected_layer_name):
        """Removes invalid characters from selected layer name to suggest a filename.

        Parameters
        ----------
        selected_layer_name : str
            The selected napari layer name.

        Returns
        -------
        suggested_name : str
            Suggested name from input selected layer name, without invalid characters.
        """
        unprintable_ascii_chars = (
            '\x00',
            '\x01',
            '\x02',
            '\x03',
            '\x04',
            '\x05',
            '\x06',
            '\x07',
            '\x08',
            '\x0e',
            '\x0f',
            '\x10',
            '\x11',
            '\x12',
            '\x13',
            '\x14',
            '\x15',
            '\x16',
            '\x17',
            '\x18',
            '\x19',
            '\x1a',
            '\x1b',
            '\x1c',
            '\x1d',
            '\x1e',
            '\x1f',
            '\x7f',
        )
        invalid_characters = (
            ''.join(unprintable_ascii_chars)
            + '/'
            + '\\'  # invalid Windows filename character
            + ':*?"<>|\t\n\r\x0b\x0c'  # invalid Windows path characters
        )
        translation_table = dict.fromkeys(map(ord, invalid_characters), None)
        # Remove invalid characters
        suggested_name = selected_layer_name.translate(translation_table)
        return suggested_name

    def _save_layers_dialog(self, selected=False):
        """Save layers (all or selected) to disk, using ``LayerList.save()``.

        Parameters
        ----------
        selected : bool
            If True, only layers that are selected in the viewer will be saved.
            By default, all layers are saved.
        """
        msg = ''
        if not len(self.viewer.layers):
            msg = trans._('There are no layers in the viewer to save')
        elif selected and not len(self.viewer.layers.selection):
            msg = trans._(
                'Please select one or more layers to save,'
                '\nor use "Save all layers..."'
            )
        if msg:
            raise OSError(trans._('Nothing to save'))

        # prepare list of extensions for drop down menu.
        ext_str, writers = _extension_string_for_layers(
            list(self.viewer.layers.selection)
            if selected
            else self.viewer.layers
        )

        msg = trans._('selected') if selected else trans._('all')
        dlg = QFileDialog()
        hist = get_save_history()
        dlg.setHistory(hist)
        # get the layer's name to use for a default name if only one layer is selected
        selected_layer_name = ''
        if self.viewer.layers.selection.active is not None:
            selected_layer_name = self.viewer.layers.selection.active.name
            selected_layer_name = self._remove_invalid_chars(
                selected_layer_name
            )
        filename, selected_filter = dlg.getSaveFileName(
            self,  # parent
            trans._('Save {msg} layers', msg=msg),  # caption
            # home dir by default if selected all, home dir and file name if only 1 layer
            str(
                Path(hist[0]) / selected_layer_name
            ),  # directory in PyQt, dir in PySide
            filter=ext_str,
            options=(
                QFileDialog.DontUseNativeDialog
                if in_ipython()
                else QFileDialog.Options()
            ),
        )
        logging.debug(
            trans._(
                'QFileDialog - filename: {filename} '
                'selected_filter: {selected_filter}',
                filename=filename or None,
                selected_filter=selected_filter or None,
            )
        )

        if filename:
            writer = _npe2_decode_selected_filter(
                ext_str, selected_filter, writers
            )
            with warnings.catch_warnings(record=True) as wa:
                saved = self.viewer.layers.save(
                    filename, selected=selected, _writer=writer
                )
                logging.debug('Saved %s', saved)
                error_messages = '\n'.join(str(x.message.args[0]) for x in wa)

            if not saved:
                raise OSError(
                    trans._(
                        'File {filename} save failed.\n{error_messages}',
                        deferred=True,
                        filename=filename,
                        error_messages=error_messages,
                    )
                )

            update_save_history(saved[0])

    def _update_welcome_screen(self):
        """Update welcome screen display based on layer count."""
        if self._show_welcome_screen:
            self._welcome_widget.set_welcome_visible(not self.viewer.layers)

    def _screenshot(self, flash=True):
        """Capture a screenshot of the Vispy canvas.

        Parameters
        ----------
        flash : bool
            Flag to indicate whether flash animation should be shown after
            the screenshot was captured.
        """
        img = self.canvas.screenshot()
        if flash:
            from napari._qt.utils import add_flash_animation

            # Here we are actually applying the effect to the `_welcome_widget`
            # and not # the `native` widget because it does not work on the
            # `native` widget. It's probably because the widget is in a stack
            # with the `QtWelcomeWidget`.
            add_flash_animation(self._welcome_widget)
        return img

    def screenshot(self, path=None, flash=True) -> np.ndarray:
        """Take currently displayed screen and convert to an image array.

        Parameters
        ----------
        path : str
            Filename for saving screenshot image.
        flash : bool
            Flag to indicate whether flash animation should be shown after
            the screenshot was captured.

        Returns
        -------
        image : array
            Numpy array of type ubyte and shape (h, w, 4). Index [0, 0] is the
            upper-left corner of the rendered region.
        """

        img = QImg2array(self._screenshot(flash))
        if path is not None:
            imsave(path, img)  # scikit-image imsave method
        return img

    def clipboard(self, flash=True):
        """Take a screenshot of the currently displayed screen and copy the
        image to the clipboard.

        Parameters
        ----------
        flash : bool
            Flag to indicate whether flash animation should be shown after
            the screenshot was captured.
        """
        cb = QGuiApplication.clipboard()
        cb.setImage(self._screenshot(flash))

    def _screenshot_dialog(self):
        """Save screenshot of current display, default .png"""
        hist = get_save_history()
        dial = ScreenshotDialog(self.screenshot, self, hist[0], hist)
        if dial.exec_():
            update_save_history(dial.selectedFiles()[0])

    def _open_file_dialog_uni(self, caption: str) -> list[str]:
        """
        Open dialog to get list of files from user
        """
        dlg = QFileDialog()
        hist = get_open_history()
        dlg.setHistory(hist)

        open_kwargs = {
            'parent': self,
            'caption': caption,
        }
        if 'pyside' in QFileDialog.__module__.lower():
            # PySide6
            open_kwargs['dir'] = hist[0]
        else:
            open_kwargs['directory'] = hist[0]

        if in_ipython():
            open_kwargs['options'] = QFileDialog.DontUseNativeDialog

        return dlg.getOpenFileNames(**open_kwargs)[0]

    def _open_files_dialog(self, choose_plugin=False, stack=False):
        """Add files from the menubar."""
        filenames = self._open_file_dialog_uni(trans._('Select file(s)...'))

        if filenames:
            self._qt_open(filenames, choose_plugin=choose_plugin, stack=stack)
            update_open_history(filenames[0])

    def _open_files_dialog_as_stack_dialog(self, choose_plugin=False):
        """Add files as a stack, from the menubar."""
        return self._open_files_dialog(choose_plugin=choose_plugin, stack=True)

    def _open_folder_dialog(self, choose_plugin=False):
        """Add a folder of files from the menubar."""
        dlg = QFileDialog()
        hist = get_open_history()
        dlg.setHistory(hist)

        folder = dlg.getExistingDirectory(
            self,
            trans._('Select folder...'),
            hist[0],  # home dir by default
            (
                QFileDialog.DontUseNativeDialog
                if in_ipython()
                else QFileDialog.Options()
            ),
        )

        if folder not in {'', None}:
            self._qt_open([folder], stack=False, choose_plugin=choose_plugin)
            update_open_history(folder)

    def _qt_open(
        self,
        filenames: list[str],
        stack: Union[bool, list[list[str]]],
        choose_plugin: bool = False,
        plugin: Optional[str] = None,
        layer_type: Optional[str] = None,
        **kwargs,
    ):
        """Open files, potentially popping reader dialog for plugin selection.

        Call ViewerModel.open and catch errors that could
        be fixed by user making a plugin choice.

        Parameters
        ----------
        filenames : List[str]
            paths to open
        choose_plugin : bool
            True if user wants to explicitly choose the plugin else False
        stack : bool or list[list[str]]
            whether to stack files or not. Can also be a list containing
            files to stack.
        plugin : str
            plugin to use for reading
        layer_type : str
            layer type for opened layers
        """
        if choose_plugin:
            handle_gui_reading(
                filenames, self, stack, plugin_override=choose_plugin, **kwargs
            )
            return

        try:
            self.viewer.open(
                filenames,
                stack=stack,
                plugin=plugin,
                layer_type=layer_type,
                **kwargs,
            )
        except ReaderPluginError as e:
            handle_gui_reading(
                filenames,
                self,
                stack,
                e.reader_plugin,
                e,
                layer_type=layer_type,
                **kwargs,
            )
        except MultipleReaderError:
            handle_gui_reading(filenames, self, stack, **kwargs)

    def toggle_console_visibility(self, event=None):
        """Toggle console visible and not visible.

        Imports the console the first time it is requested.
        """
        if in_ipython() or in_jupyter():
            return

        # force instantiation of console if not already instantiated
        _ = self.console

        viz = not self.dockConsole.isVisible()
        # modulate visibility at the dock widget level as console is dockable
        self.dockConsole.setVisible(viz)
        if self.dockConsole.isFloating():
            self.dockConsole.setFloating(True)

        if viz:
            self.dockConsole.raise_()
            self.dockConsole.setFocus()

        self.viewerButtons.consoleButton.setProperty(
            'expanded', self.dockConsole.isVisible()
        )
        self.viewerButtons.consoleButton.style().unpolish(
            self.viewerButtons.consoleButton
        )
        self.viewerButtons.consoleButton.style().polish(
            self.viewerButtons.consoleButton
        )

    def set_welcome_visible(self, visible):
        """Show welcome screen widget."""
        self._show_welcome_screen = visible
        self._welcome_widget.set_welcome_visible(visible)

    def keyPressEvent(self, event):
        """Called whenever a key is pressed.

        Parameters
        ----------
        event : qtpy.QtCore.QEvent
            Event from the Qt context.
        """
        self.canvas._scene_canvas._backend._keyEvent(
            self.canvas._scene_canvas.events.key_press, event
        )
        event.accept()

    def keyReleaseEvent(self, event):
        """Called whenever a key is released.

        Parameters
        ----------
        event : qtpy.QtCore.QEvent
            Event from the Qt context.
        """
        self.canvas._scene_canvas._backend._keyEvent(
            self.canvas._scene_canvas.events.key_release, event
        )
        event.accept()

    def dragEnterEvent(self, event):
        """Ignore event if not dragging & dropping a file or URL to open.

        Using event.ignore() here allows the event to pass through the
        parent widget to its child widget, otherwise the parent widget
        would catch the event and not pass it on to the child widget.

        Parameters
        ----------
        event : qtpy.QtCore.QDragEvent
            Event from the Qt context.
        """
        if event.mimeData().hasUrls():
            self._set_drag_status()
            event.accept()
        else:
            event.ignore()

    def _set_drag_status(self):
        """Set dedicated status message when dragging files into viewer"""
        self.viewer.status = trans._(
            'Hold <Alt> key to open plugin selection. Hold <Shift> to open files as stack.'
        )

    def _image_from_clipboard(self):
        """Insert image from clipboard as a new layer if clipboard contains an image or link."""
        cb = QGuiApplication.clipboard()
        if cb.mimeData().hasImage():
            image = cb.image()
            if image.isNull():
                return
            arr = QImg2array(image)
            self.viewer.add_image(arr)
            return
        if cb.mimeData().hasUrls():
            show_info('No image in clipboard, trying to open link instead.')
            self._open_from_list_of_urls_data(
                cb.mimeData().urls(), stack=False, choose_plugin=False
            )
            return
        if cb.mimeData().hasText():
            show_info(
                'No image in clipboard, trying to parse text in clipboard as a link.'
            )
            url_list = []
            for line in cb.mimeData().text().split('\n'):
                url = QUrl(line.strip())
                if url.isEmpty():
                    continue
                if url.scheme() == '':
                    url.setScheme('file')
                if url.isLocalFile() and not Path(url.toLocalFile()).exists():
                    break
                url_list.append(url)
            else:
                self._open_from_list_of_urls_data(
                    url_list, stack=False, choose_plugin=False
                )
                return
        show_info('No image or link in clipboard.')

    def dropEvent(self, event):
        """Add local files and web URLS with drag and drop.

        For each file, attempt to open with existing associated reader
        (if available). If no reader is associated or opening fails,
        and more than one reader is available, open dialog and ask
        user to choose among available readers. User can choose to persist
        this choice.

        Parameters
        ----------
        event : qtpy.QtCore.QDropEvent
            Event from the Qt context.
        """
        shift_down = (
            QGuiApplication.keyboardModifiers()
            & Qt.KeyboardModifier.ShiftModifier
        )
        alt_down = (
            QGuiApplication.keyboardModifiers()
            & Qt.KeyboardModifier.AltModifier
        )
        self._open_from_list_of_urls_data(
            event.mimeData().urls(),
            stack=bool(shift_down),
            choose_plugin=bool(alt_down),
        )

    def _open_from_list_of_urls_data(
        self, urls_list: list[QUrl], stack: bool, choose_plugin: bool
    ):
        filenames = []
        for url in urls_list:
            if url.isLocalFile():
                # directories get a trailing "/", Path conversion removes it
                filenames.append(str(Path(url.toLocalFile())))
            else:
                filenames.append(url.toString())

        self._qt_open(
            filenames,
            stack=stack,
            choose_plugin=choose_plugin,
        )

    def closeEvent(self, event):
        """Cleanup and close.

        Parameters
        ----------
        event : qtpy.QtCore.QCloseEvent
            Event from the Qt context.
        """
        if self._layers is not None:
            # do not create layerlist if it does not exist yet.
            self.layers.close()

        # if the viewer.QtDims object is playing an axis, we need to terminate
        # the AnimationThread before close, otherwise it will cause a segFault
        # or Abort trap. (calling stop() when no animation is occurring is also
        # not a problem)
        self.dims.stop()
        self.canvas.delete()
        if self._console is not None:
            self.console.close()
        self.dockConsole.deleteLater()
        event.accept()


if TYPE_CHECKING:
    from napari._qt.experimental.qt_poll import QtPoll
    from napari.components.experimental.remote import RemoteManager


def _create_qt_poll(parent: QObject, camera: Camera) -> Optional[QtPoll]:
    """Create and return a QtPoll instance, if needed.

    Create a QtPoll instance for the monitor.

    Monitor needs QtPoll to poll for incoming messages. This might be
    temporary until we can process incoming messages with a dedicated
    thread.

    Parameters
    ----------
    parent : QObject
        Parent Qt object.
    camera : Camera
        Camera that the QtPoll object will listen to.

    Returns
    -------
    Optional[QtPoll]
        The new QtPoll instance, if we need one.
    """
    if not config.monitor:
        return None

    from napari._qt.experimental.qt_poll import QtPoll

    qt_poll = QtPoll(parent)
    camera.events.connect(qt_poll.on_camera)
    return qt_poll


def _create_remote_manager(
    layers: LayerList, qt_poll
) -> Optional[RemoteManager]:
    """Create and return a RemoteManager instance, if we need one.

    Parameters
    ----------
    layers : LayersList
        The viewer's layers.
    qt_poll : QtPoll
        The viewer's QtPoll instance.
    """
    if not config.monitor:
        return None  # Not using the monitor at all

    from napari.components.experimental.monitor import monitor
    from napari.components.experimental.remote import RemoteManager

    # Start the monitor so we can access its events. The monitor has no
    # dependencies to napari except to utils.Event.
    started = monitor.start()

    if not started:
        return None  # Probably not >= Python 3.9, so no manager is needed.

    # Create the remote manager and have monitor call its process_command()
    # method to execute commands from clients.
    manager = RemoteManager(layers)

    # RemoteManager will process incoming command from the monitor.
    monitor.run_command_event.connect(manager.process_command)

    # QtPoll should pool the RemoteManager and the Monitor.
    qt_poll.events.poll.connect(manager.on_poll)
    qt_poll.events.poll.connect(monitor.on_poll)

    return manager


def _in_napari(n: int, frame: FrameType):
    """
    Determines whether we are in napari by looking at:
        1) the frames modules names:
        2) the min_depth
    """
    if n < 2:
        return True
    # in-n-out is used in napari for dependency injection.
    for pref in {'napari.', 'in_n_out.'}:
        if frame.f_globals.get('__name__', '').startswith(pref):
            return True
    return False



from unittest import mock

import numpy as np
import pytest
from app_model.types import MenuItem, SubmenuItem
from npe2 import DynamicPlugin
from npe2.manifest.contributions import SampleDataURI
from qtpy.QtGui import QGuiApplication

from napari._app_model import get_app_model
from napari._app_model.constants import MenuId
from napari._qt._qapp_model._tests.utils import get_submenu_action
from napari.layers import Image
from napari.plugins._tests.test_npe2 import mock_pm  # noqa: F401
from napari.utils.action_manager import action_manager


def test_sample_data_triggers_reader_dialog(
    make_napari_viewer, tmp_plugin: DynamicPlugin
):
    """Sample data pops reader dialog if multiple compatible readers"""
    # make two tmp readers that take tif files
    tmp2 = tmp_plugin.spawn(register=True)

    @tmp_plugin.contribute.reader(filename_patterns=['*.tif'])
    def _(path): ...

    @tmp2.contribute.reader(filename_patterns=['*.tif'])
    def _(path): ...

    # make a sample data reader for tif file
    my_sample = SampleDataURI(
        key='tmp-sample',
        display_name='Temp Sample',
        uri='some-path/some-file.tif',
    )
    tmp_plugin.manifest.contributions.sample_data = [my_sample]
    app = get_app_model()
    # Configures `app`, registers actions and initializes plugins
    make_napari_viewer()
    with mock.patch(
        'napari._qt.dialogs.qt_reader_dialog.handle_gui_reading'
    ) as mock_read:
        app.commands.execute_command('tmp_plugin:tmp-sample')

    # assert that handle gui reading was called
    mock_read.assert_called_once()


def test_plugin_display_name_use_for_multiple_samples(
    make_napari_viewer,
    builtins,
):
    """Check 'display_name' used for submenu when plugin has >1 sample data."""
    app = get_app_model()
    viewer = make_napari_viewer()

    # builtins provides more than one sample,
    # so the submenu should use the `display_name` from manifest
    samples_menu = app.menus.get_menu(MenuId.FILE_SAMPLES)
    assert samples_menu[0].title == 'napari builtins'
    # Now ensure that the actions are still correct
    # trigger the action, opening the first sample: `Astronaut`
    assert 'napari:astronaut' in app.commands
    assert len(viewer.layers) == 0
    app.commands.execute_command('napari:astronaut')
    assert len(viewer.layers) == 1
    assert viewer.layers[0].name == 'astronaut'


def test_sample_menu_plugin_state_change(
    make_napari_viewer,
    tmp_plugin: DynamicPlugin,
):
    """Check samples submenu correct after plugin changes state."""

    app = get_app_model()
    pm = tmp_plugin.plugin_manager
    # Check no samples menu before plugin registration
    with pytest.raises(KeyError):
        app.menus.get_menu(MenuId.FILE_SAMPLES)

    sample1 = SampleDataURI(
        key='tmp-sample-1',
        display_name='Temp Sample One',
        uri='some-file.tif',
    )
    sample2 = SampleDataURI(
        key='tmp-sample-2',
        display_name='Temp Sample Two',
        uri='some-file.tif',
    )
    tmp_plugin.manifest.contributions.sample_data = [sample1, sample2]

    # Configures `app`, registers actions and initializes plugins
    make_napari_viewer()

    samples_menu = app.menus.get_menu(MenuId.FILE_SAMPLES)
    assert len(samples_menu) == 1
    assert isinstance(samples_menu[0], SubmenuItem)
    assert samples_menu[0].title == tmp_plugin.display_name
    samples_sub_menu = app.menus.get_menu(MenuId.FILE_SAMPLES + '/tmp_plugin')
    assert len(samples_sub_menu) == 2
    assert isinstance(samples_sub_menu[0], MenuItem)
    assert samples_sub_menu[0].command.title == 'Temp Sample One'
    assert 'tmp_plugin:tmp-sample-1' in app.commands

    # Disable plugin
    pm.disable(tmp_plugin.name)
    with pytest.raises(KeyError):
        app.menus.get_menu(MenuId.FILE_SAMPLES)
    assert 'tmp_plugin:tmp-sample-1' not in app.commands

    # Enable plugin
    pm.enable(tmp_plugin.name)
    samples_sub_menu = app.menus.get_menu(MenuId.FILE_SAMPLES + '/tmp_plugin')
    assert len(samples_sub_menu) == 2
    assert 'tmp_plugin:tmp-sample-1' in app.commands


def test_sample_menu_single_data(
    make_napari_viewer,
    tmp_plugin: DynamicPlugin,
):
    """Checks sample submenu correct when plugin has single sample data."""
    app = get_app_model()
    sample = SampleDataURI(
        key='tmp-sample-1',
        display_name='Temp Sample One',
        uri='some-file.tif',
    )
    tmp_plugin.manifest.contributions.sample_data = [sample]
    # Configures `app`, registers actions and initializes plugins
    make_napari_viewer()

    samples_menu = app.menus.get_menu(MenuId.FILE_SAMPLES)
    assert isinstance(samples_menu[0], MenuItem)
    assert len(samples_menu) == 1
    assert samples_menu[0].command.title == 'Temp Sample One (Temp Plugin)'
    assert 'tmp_plugin:tmp-sample-1' in app.commands


def test_sample_menu_sorted(
    mock_pm,  # noqa: F811
    mock_app_model,
    tmp_plugin: DynamicPlugin,
):
    from napari._app_model import get_app_model
    from napari.plugins import _initialize_plugins

    # we make sure 'plugin-b' is registered first
    tmp_plugin2 = tmp_plugin.spawn(name='plugin-b', register=True)
    tmp_plugin1 = tmp_plugin.spawn(name='plugin-a', register=True)

    @tmp_plugin1.contribute.sample_data(display_name='Sample 1')
    def sample1(): ...

    @tmp_plugin1.contribute.sample_data(display_name='Sample 2')
    def sample2(): ...

    @tmp_plugin2.contribute.sample_data(display_name='Sample 1')
    def sample2_1(): ...

    @tmp_plugin2.contribute.sample_data(display_name='Sample 2')
    def sample2_2(): ...

    _initialize_plugins()
    samples_menu = list(get_app_model().menus.get_menu('napari/file/samples'))
    submenus = [item for item in samples_menu if isinstance(item, SubmenuItem)]
    assert len(submenus) == 3
    # mock_pm registers a sample_manifest with two sample data contributions
    assert submenus[0].title == 'My Plugin'
    assert submenus[1].title == 'plugin-a'
    assert submenus[2].title == 'plugin-b'


def test_show_shortcuts_actions(make_napari_viewer):
    viewer = make_napari_viewer()
    assert viewer.window._pref_dialog is None
    action_manager.trigger('napari:show_shortcuts')
    assert viewer.window._pref_dialog is not None
    assert viewer.window._pref_dialog._list.currentItem().text() == 'Shortcuts'
    viewer.window._pref_dialog.close()


def test_image_from_clipboard(make_napari_viewer):
    make_napari_viewer()
    app = get_app_model()

    # Ensure clipboard is empty
    QGuiApplication.clipboard().clear()
    clipboard_image = QGuiApplication.clipboard().image()
    assert clipboard_image.isNull()

    # Check action command execution
    with mock.patch('napari._qt.qt_viewer.show_info') as mock_show_info:
        app.commands.execute_command(
            'napari.window.file._image_from_clipboard'
        )
    mock_show_info.assert_called_once_with('No image or link in clipboard.')


@pytest.mark.parametrize(
    ('action_id', 'dialog_method', 'dialog_return', 'filename_call', 'stack'),
    [
        (
            # Open File(s)...
            'napari.window.file.open_files_dialog',
            'getOpenFileNames',
            (['my-file.tif'], ''),
            ['my-file.tif'],
            False,
        ),
        (
            # Open Files as Stack...
            'napari.window.file.open_files_as_stack_dialog',
            'getOpenFileNames',
            (['my-file.tif'], ''),
            ['my-file.tif'],
            True,
        ),
        (
            # Open Folder...
            'napari.window.file.open_folder_dialog',
            'getExistingDirectory',
            'my-dir/',
            ['my-dir/'],
            False,
        ),
    ],
)
def test_open(
    make_napari_viewer,
    action_id,
    dialog_method,
    dialog_return,
    filename_call,
    stack,
):
    """Test base `Open ...` actions can be triggered."""
    make_napari_viewer()
    app = get_app_model()

    # Check action command execution
    with (
        mock.patch('napari._qt.qt_viewer.QFileDialog') as mock_file,
        mock.patch('napari._qt.qt_viewer.QtViewer._qt_open') as mock_read,
    ):
        mock_file_instance = mock_file.return_value
        getattr(mock_file_instance, dialog_method).return_value = dialog_return
        app.commands.execute_command(action_id)
    mock_read.assert_called_once_with(
        filename_call, stack=stack, choose_plugin=False
    )


@pytest.mark.parametrize(
    (
        'menu_str',
        'dialog_method',
        'dialog_return',
        'filename_call',
        'stack',
    ),
    [
        (
            'Open File(s)...',
            'getOpenFileNames',
            (['my-file.tif'], ''),
            ['my-file.tif'],
            False,
        ),
        (
            'Open Files as Stack...',
            'getOpenFileNames',
            (['my-file.tif'], ''),
            ['my-file.tif'],
            True,
        ),
        (
            'Open Folder...',
            'getExistingDirectory',
            'my-dir/',
            ['my-dir/'],
            False,
        ),
    ],
)
def test_open_with_plugin(
    make_napari_viewer,
    menu_str,
    dialog_method,
    dialog_return,
    filename_call,
    stack,
):
    viewer = make_napari_viewer()
    action, _a = get_submenu_action(
        viewer.window.file_menu, 'Open with Plugin', menu_str
    )
    with (
        mock.patch('napari._qt.qt_viewer.QFileDialog') as mock_file,
        mock.patch('napari._qt.qt_viewer.QtViewer._qt_open') as mock_read,
    ):
        mock_file_instance = mock_file.return_value
        getattr(mock_file_instance, dialog_method).return_value = dialog_return
        action.trigger()
    mock_read.assert_called_once_with(
        filename_call, stack=stack, choose_plugin=True
    )


def test_preference_dialog(make_napari_viewer):
    """Test preferences action can be triggered."""
    make_napari_viewer()
    app = get_app_model()

    # Check action command execution
    with (
        mock.patch(
            'napari._qt.qt_main_window.PreferencesDialog.show'
        ) as mock_pref_dialog_show,
    ):
        app.commands.execute_command(
            'napari.window.file.show_preferences_dialog'
        )
    mock_pref_dialog_show.assert_called_once()


def test_save_layers_enablement_updated_context(make_napari_viewer, builtins):
    """Test that enablement status of save layer actions updated correctly."""
    get_app_model()
    viewer = make_napari_viewer()

    save_layers_action = viewer.window.file_menu.findAction(
        'napari.window.file.save_layers_dialog',
    )
    save_selected_layers_action = viewer.window.file_menu.findAction(
        'napari.window.file.save_layers_dialog.selected',
    )
    # Check both save actions are not enabled when no layers
    assert len(viewer.layers) == 0
    viewer.window._update_file_menu_state()
    assert not save_layers_action.isEnabled()
    assert not save_selected_layers_action.isEnabled()

    # Add selected layer and check both save actions enabled
    layer = Image(np.random.random((10, 10)))
    viewer.layers.append(layer)
    assert len(viewer.layers) == 1
    viewer.window._update_file_menu_state()
    assert save_layers_action.isEnabled()
    assert save_selected_layers_action.isEnabled()

    # Remove selection and check 'Save All Layers...' is enabled but
    # 'Save Selected Layers...' is not
    viewer.layers.selection.clear()
    viewer.window._update_file_menu_state()
    assert save_layers_action.isEnabled()
    assert not save_selected_layers_action.isEnabled()


@pytest.mark.parametrize(
    ('action_id', 'dialog_method', 'dialog_return'),
    [
        (
            # Save Selected Layers...
            'napari.window.file.save_layers_dialog.selected',
            'getSaveFileName',
            (None, None),
        ),
        (
            # Save All Layers...
            'napari.window.file.save_layers_dialog',
            'getSaveFileName',
            (None, None),
        ),
    ],
)
def test_save_layers(
    make_napari_viewer, action_id, dialog_method, dialog_return
):
    """Test save layer selected/all actions can be triggered."""
    viewer = make_napari_viewer()
    app = get_app_model()

    # Add selected layer
    layer = Image(np.random.random((10, 10)))
    viewer.layers.append(layer)
    assert len(viewer.layers) == 1
    viewer.window._update_file_menu_state()

    # Check action command execution
    with mock.patch('napari._qt.qt_viewer.QFileDialog') as mock_file:
        mock_file_instance = mock_file.return_value
        getattr(mock_file_instance, dialog_method).return_value = dialog_return
        app.commands.execute_command(action_id)
    mock_file.assert_called_once()


@pytest.mark.parametrize(
    ('action_id', 'patch_method', 'dialog_return'),
    [
        (
            # Save Screenshot with Viewer...
            'napari.window.file.save_viewer_screenshot_dialog',
            'napari._qt.dialogs.screenshot_dialog.ScreenshotDialog.exec_',
            False,
        ),
    ],
)
def test_screenshot(
    make_napari_viewer, action_id, patch_method, dialog_return
):
    """Test screenshot actions can be triggered."""
    make_napari_viewer()
    app = get_app_model()

    # Check action command execution
    with mock.patch(patch_method) as mock_screenshot:
        mock_screenshot.return_value = dialog_return
        app.commands.execute_command(action_id)
    mock_screenshot.assert_called_once()


@pytest.mark.parametrize(
    'action_id',
    [
        # Copy Screenshot with Viewer to Clipboard
        'napari.window.file.copy_viewer_screenshot',
    ],
)
def test_screenshot_to_clipboard(make_napari_viewer, qtbot, action_id):
    """Test screenshot to clipboard actions can be triggered."""
    viewer = make_napari_viewer()
    app = get_app_model()

    # Add selected layer
    layer = Image(np.random.random((10, 10)))
    viewer.layers.append(layer)
    assert len(viewer.layers) == 1
    viewer.window._update_file_menu_state()

    # Check action command execution
    # ---- Ensure clipboard is empty
    QGuiApplication.clipboard().clear()
    clipboard_image = QGuiApplication.clipboard().image()
    assert clipboard_image.isNull()
    # ---- Execute action
    with mock.patch('napari._qt.utils.add_flash_animation') as mock_flash:
        app.commands.execute_command(action_id)
    mock_flash.assert_called_once()
    # ---- Ensure clipboard has image
    clipboard_image = QGuiApplication.clipboard().image()
    assert not clipboard_image.isNull()


@pytest.mark.parametrize(
    (
        'action_id',
        'patch_method',
    ),
    [
        (
            # Restart
            'napari.window.file.restart',
            'napari._qt.qt_main_window._QtMainWindow.restart',
        ),
    ],
)
def test_restart(make_napari_viewer, action_id, patch_method):
    """Testrestart action can be triggered."""
    make_napari_viewer()
    app = get_app_model()

    # Check action command execution
    with mock.patch(patch_method) as mock_restart:
        app.commands.execute_command(action_id)
    mock_restart.assert_called_once()


@pytest.mark.parametrize(
    ('action_id', 'patch_method', 'method_params'),
    [
        (
            # Close Window
            'napari.window.file.close_dialog',
            'napari._qt.qt_main_window._QtMainWindow.close',
            (False, True),
        ),
        (
            # Exit
            'napari.window.file.quit_dialog',
            'napari._qt.qt_main_window._QtMainWindow.close',
            (True, True),
        ),
    ],
)
def test_close(make_napari_viewer, action_id, patch_method, method_params):
    """Test close/exit actions can be triggered."""
    make_napari_viewer()
    app = get_app_model()
    quit_app, confirm_need = method_params

    # Check action command execution
    with mock.patch(patch_method) as mock_close:
        app.commands.execute_command(action_id)
    mock_close.assert_called_once_with(
        quit_app=quit_app, confirm_need=confirm_need
    )
